"""
Functions to load JSON or XML files for conversion into authorityspoke objects.

Will usually hand off data to the io.readers module to create authorityspoke objects.
"""
import json
import pathlib

from typing import Any, Dict, List, Iterator, Optional, Tuple, Union

import yaml

from legislice.download import Client


from authorityspoke.decisions import Decision
from authorityspoke.holdings import Holding
from authorityspoke.opinions import AnchoredHoldings
from authorityspoke.rules import Rule

from authorityspoke.io import filepaths, readers
from authorityspoke.io.name_index import Mentioned
from authorityspoke.io.schemas_yaml import (
    RawEnactment,
    RawFactor,
    RawHolding,
    RawDecision,
)


def load_anchored_holdings(
    filename: Optional[str] = None,
    directory: Optional[pathlib.Path] = None,
    filepath: Optional[pathlib.Path] = None,
) -> Dict[str, Union[Dict[str, RawFactor], Dict[str, RawEnactment], List[RawHolding]]]:
    r"""
    Load list of records from YAML or JSON to create :class:`.Holding`\s with text selectors.

    :param filename:
        the name of the JSON file to look in for :class:`Holding`
        data in the format that lists ``mentioned_factors``
        followed by a list of holdings.

    :param directory:
        the path of the directory containing the JSON file.

    :param filepath:
        Complete path to the XML file representing the :class:`.Code`,
        including filename.

    :returns:
        a list of :class:`Holding`\s from a JSON file in the
        ``example_data/holdings`` subdirectory, from a JSON
        file.
    """
    validated_filepath = filepaths.make_filepath(
        filename, directory, filepath, default_folder="holdings"
    )

    with open(validated_filepath, "r") as f:
        if validated_filepath.suffix == ".yaml":
            holdings = yaml.safe_load(f)
        else:
            holdings = json.load(f)

    return holdings


def load_holdings(
    filename: Optional[str] = None,
    directory: Optional[pathlib.Path] = None,
    filepath: Optional[pathlib.Path] = None,
) -> Dict[str, Union[Dict[str, RawFactor], Dict[str, RawEnactment], List[RawHolding]]]:
    r"""
    Load a list of records from JSON to create :class:`.Holding`\s.

    :param filename:
        the name of the JSON file to look in for :class:`Holding`
        data in the format that lists ``mentioned_factors``
        followed by a list of holdings.

    :param directory:
        the path of the directory containing the JSON file.

    :param filepath:
        Complete path to the XML file representing the :class:`.Code`,
        including filename.

    :returns:
        a list of :class:`Holding`\s from a JSON file in the
        ``example_data/holdings`` subdirectory, from a JSON
        file.
    """
    holdings = load_anchored_holdings(
        filename=filename, directory=directory, filepath=filepath
    )

    if isinstance(holdings, Dict):
        return holdings["holdings"]
    return holdings


def load_rules_with_index(
    filename: Optional[str] = None,
    directory: Optional[pathlib.Path] = None,
    filepath: Optional[pathlib.Path] = None,
    client: Optional[Client] = None,
    many: bool = True,
) -> Tuple[List[Rule], Mentioned]:
    r"""
    Read :class:`.Rule`\s from a file.

    Even though this function will generate :class:`.Rule`\s instead
    of :class:`.Holding`\s, it uses the :func:`load_holding` function.
    Some fields that can exist in a :class:`.Holding` object (rule_valid,
    decided, exclusive) are not applicable to :class:`.Rule`\s.

    :param filename: The name of the input JSON file.

    :param directory: The directory where the input JSON file is located.

    :param filepath:
        Complete path to the JSON file representing the :class:`.Opinion`,
        including filename.

    :param regime:
        The regime to reference for the :class:`Enactment`\s
        mentioned in the holding.
    """
    raw_rules = load_holdings(filename=filename, directory=directory, filepath=filepath)
    return readers.read_rules_with_index(raw_rules, client=client, many=many)


def read_holdings_from_file(
    filename: Optional[str] = None,
    directory: Optional[pathlib.Path] = None,
    filepath: Optional[pathlib.Path] = None,
    client: Optional[Client] = None,
) -> List[Holding]:
    r"""
    Read holdings from a file.

    :param filename: The name of the input JSON file.

    :param directory: The directory where the input JSON file is located.

    :param filepath:
        Complete path to the JSON file representing the :class:`.Opinion`,
        including filename.

    :param client:
        The client with an API key to download :class:`Enactment`\s
        mentioned in the holding.
    """
    raw_holdings = load_holdings(
        filename=filename, directory=directory, filepath=filepath
    )

    return readers.read_holdings(raw_holdings, client=client)


def read_anchored_holdings_from_file(
    filename: Optional[str] = None,
    directory: Optional[pathlib.Path] = None,
    filepath: Optional[pathlib.Path] = None,
    client: Optional[Client] = None,
) -> AnchoredHoldings:
    r"""
    Read holdings from file, with Opinion text anchors for holdings and factors.

    This function can accept a file containing :class:`.authorityspoke.holdings.Holding`
    summaries in the YAML format
    that may contain abbreviations and expandable named entities.

    In the example below, a :class:`~authorityspoke.io.fake_enactments.FakeClient` is used to
    add fields to :class:`~legislice.enactments.Enactment`\s in the Holding
    objects, to avoid using the network or making API calls. The
    real :class:`~authorityspoke.io.downloads.LegisClient` class
    would also have worked (with an appropriate API key).

        >>> from authorityspoke.io.fake_enactments import FakeClient
        >>> fake_client = FakeClient.from_file("usc.json")
        >>> filepath = filepaths.make_filepath(filename="holding_mazza_alaluf.yaml")
        >>> with open(filepath, "r") as f:
        ...     yaml.safe_load(f)
        {'holdings': [{'inputs': [{'type': 'fact', 'content': "{Mazza-Alaluf} used Mazza-Alaluf's business {Turismo Costa Brava} to commit the New York offense of engaging in the business of receiving money for transmission or transmitting the same, without a license therefor"}], 'outputs': [{'type': 'fact', 'content': 'Mazza-Alaluf operated Turismo Costa Brava without an appropriate money transmitting license in a State where such operation was punishable as a misdemeanor or a felony under State law', 'anchors': "we conclude that sufficient evidence supports Mazza-Alaluf's convictions under 18 U.S.C. § 1960(b)(1)(A) for conspiring to operate and operating a money transmitting business without appropriate state licenses.", 'name': 'operated without license'}], 'enactments': [{'node': '/us/usc/t18/s1960/b/1', 'anchors': 'state money transmitting licenses, see |18 U.S.C. § 1960(b)(1)(A)|', 'exact': 'is operated without an appropriate money transmitting license in a State where such operation is punishable as a misdemeanor or a felony under State law, whether or not the defendant knew that the operation was required to be licensed or that the operation was so punishable', 'name': 'state money transmitting license provision'}], 'universal': True}, {'inputs': ['operated without license', {'type': 'fact', 'content': 'Mazza-Alaluf operated Turismo Costa Brava as a business', 'anchors': 'Mazza-Alaluf does not contest that he owned and managed Turismo'}, {'type': 'fact', 'content': 'Turismo Costa Brava was a money transmitting business', 'anchors': 'record evidence that Turismo conducted substantial money transmitting business in the three states'}], 'despite': [{'type': 'fact', 'content': 'Turismo Costa Brava was a domestic financial institution', 'truth': False, 'anchors': 'without respect to whether or not Turismo was a "domestic financial institution"'}], 'outputs': [{'type': 'fact', 'content': 'Mazza-Alaluf committed the offense of conducting a money transmitting business without a license required by state law', 'anchors': 'a crime to operate a money transmitting business without appropriate state licenses,'}], 'enactments': ['state money transmitting license provision'], 'enactments_despite': [{'node': '/us/usc/t31/s5312/b/1', 'anchors': ['§ 5312(b)(1) (defining "domestic financial institution")']}], 'anchors': [{'prefix': 'Accordingly, we conclude that the', 'suffix': 'In any event'}], 'universal': True, 'mandatory': True}]}
        >>> result = read_anchored_holdings_from_file(filepath=filepath, client=fake_client)
        >>> selector = result.named_anchors["the fact it was false that <Turismo Costa Brava> was a domestic financial institution"][0]
        >>> selector.exact
        'without respect to whether or not Turismo was a "domestic financial institution"'
        >>> print(result.holdings[0].outputs[0])
        the fact that <Mazza-Alaluf> operated <Turismo Costa Brava> without an appropriate money transmitting license in a State where such operation was punishable as a misdemeanor or a felony under State law
    """
    raw_holdings = load_anchored_holdings(
        filename=filename, directory=directory, filepath=filepath
    )
    return readers.read_holdings_with_anchors(raw_holdings, client=client)


def load_decision(
    filename: Optional[str] = None,
    directory: Optional[pathlib.Path] = None,
    filepath: Optional[pathlib.Path] = None,
) -> RawDecision:
    r"""
    Load file containing a judicial decision with one or more opinions.

    Relies on the JSON format from the `Caselaw Access Project
    API <https://api.case.law/v1/cases/>`_.

    :param filename: The name of the input JSON file.

    :param directory: The directory where the input JSON file is located.

    :param filepath:
        Complete path to the JSON file representing the :class:`.Opinion`,
        including filename.
    """

    validated_filepath = filepaths.make_filepath(
        filename, directory, filepath, default_folder="cases"
    )

    with open(validated_filepath, "r") as f:
        decision_dict = json.load(f)

    return decision_dict


def load_and_read_decision(
    filename: Optional[str] = None,
    directory: Optional[pathlib.Path] = None,
    filepath: Optional[pathlib.Path] = None,
) -> Decision:
    r"""Load file containing a judicial decision and make an Opinion object."""
    raw_decision = load_decision(
        filename=filename, directory=directory, filepath=filepath
    )
    return readers.read_decision(raw_decision)



from nettlesome.terms import ContextRegister, means, Explanation
from nettlesome.entities import Entity

from authorityspoke.io.readers import read_fact
from authorityspoke.io.text_expansion import expand_shorthand


class TestContext:
    al = expand_shorthand({"content": "{Al} sold {the bull} to {Betty}."})
    alice = expand_shorthand({"content": "{Alice} sold {the cow} to {Bob}."})

    def test_impossible_register(self):
        fact_al = read_fact(self.al)
        fact_alice = read_fact(self.alice)
        context = ContextRegister()
        context.insert_pair(Entity("Al"), Entity("Bob"))
        answers = fact_al.update_context_register(fact_alice, context, means)
        assert not any(answers)

    def test_possible_register(self):
        fact_al = read_fact(self.al)
        fact_alice = read_fact(self.alice)
        register = ContextRegister()
        register.insert_pair(Entity("Al"), Entity("Alice"))
        answers = fact_al.update_context_register(fact_alice, register, means)
        assert Entity("the bull").key in next(answers).keys()

    def test_explain_consistency(self):
        fact_al = read_fact(self.al)
        fact_alice = read_fact(self.alice)
        register = ContextRegister()
        register.insert_pair(Entity("Al"), Entity("Alice"))
        explanation = fact_al.explain_consistent_with(fact_alice, register)
        assert "<the bull> is like <the cow>" in str(explanation.context)


class TestExplainHoldings:
    def test_explain_implication(self, make_decision_with_holding):
        oracle = make_decision_with_holding["oracle"]
        left = oracle.holdings[18]
        right = oracle.holdings[19]
        explanation = left.explain_implication(right)

        assert "implies" in str(explanation).lower()



import datetime
import os

from anchorpoint.textselectors import TextQuoteSelector, TextSelectionError
from dotenv import load_dotenv
from legislice.download import Client

from legislice.citations import CodeLevel
from legislice.enactments import Enactment
from legislice.groups import EnactmentGroup
from legislice.schemas import EnactmentSchema

from nettlesome.entities import Entity
from nettlesome.predicates import Predicate
from nettlesome.statements import Statement
import pytest

from authorityspoke.io import loaders, readers, schemas_yaml, dump
from authorityspoke.io.fake_enactments import FakeClient


load_dotenv()

TOKEN = os.getenv("LEGISLICE_API_TOKEN")


class TestEnactments:
    client = Client(api_token=TOKEN)

    def test_make_enactment(self, e_search_clause):
        search_clause = e_search_clause
        assert search_clause.selected_text().endswith("shall not be violated…")

    def test_create_enactment_with_init(self, fake_beard_client):
        """
        Using the __init__ method of the Enactment class, insteaid of
        readers.read_enactment or the Enactment marshmallow schema.
        """
        beard_definition = fake_beard_client.read("/test/acts/47/4/")
        assert beard_definition.text.startswith("In this Act, beard")

    def test_make_enactment_from_selector_without_code(self, fake_usc_client):
        selector = TextQuoteSelector(suffix=" to their respective")
        art_3 = fake_usc_client.read("/us/const/article/I/8/8")
        art_3.select(selector)
        assert art_3.text.startswith("To promote")
        assert art_3.selected_text().endswith("exclusive Right…")

    def test_make_enactment_from_dict_with_reader(self, fake_usc_client):
        fourth_a = fake_usc_client.read_from_json({"node": "/us/const/amendment/IV"})
        assert fourth_a.text.endswith("and the persons or things to be seized.")

    def test_make_enactment_from_dict_with_text_split(self, fake_usc_client):
        fourth_a = fake_usc_client.read_from_json(
            {
                "node": "/us/const/amendment/IV",
                "text": "and|the persons or things|to be seized.",
            }
        )
        assert fourth_a.selected_text() == "…the persons or things…"

    def test_passage_from_imported_statute(self, fake_usc_client):
        oracle_majority = loaders.load_and_read_decision(f"oracle_h.json").majority
        loaded = loaders.load_holdings("holding_oracle.json")
        holdings = readers.read_holdings(loaded, client=fake_usc_client)
        oracle_majority.posit(holdings)
        despite_text = str(list(oracle_majority.holdings)[5])
        assert "In no case does copyright protection " in despite_text

    def test_short_passage_from_uslm_code(self, fake_usc_client):
        """Also tests adding the missing initial "/" in ``path``."""
        method = fake_usc_client.read_from_json(
            {
                "node": "us/usc/t17/s102/b",
                "prefix": "process, system,",
                "suffix": ", concept, principle",
            },
        )
        assert method.selected_text() == "…method of operation…"

    def test_chapeau_and_subsections_from_uslm_code(self, fake_beard_client):
        definition = fake_beard_client.read_from_json({"node": "/test/acts/47/4"})
        assert definition.text.strip().endswith("below the nose.")

    def test_cite_path_in_str(self, e_search_clause):
        assert "/us/const/amendment/IV" in str(e_search_clause)

    def test_unequal_to_statement(self, watt_factor, e_copyright):
        stole_predicate = Predicate("$defendant stole $object")
        stole_fact = Statement(
            stole_predicate, terms=[Entity("Alice"), Entity("the gold bar")]
        )

        assert not stole_fact.means(e_copyright)
        with pytest.raises(TypeError):
            e_copyright.means(watt_factor["f1"])

    def test_equal_enactment_text(self, e_due_process_5, e_due_process_14):
        assert e_due_process_5.means(e_due_process_14)

    def test_not_gt_if_equal(self, e_search_clause):
        assert e_search_clause == e_search_clause
        assert e_search_clause.means(e_search_clause)
        assert not e_search_clause > e_search_clause

    def test_enactment_subset_or_equal(self, e_due_process_5, e_due_process_14):
        assert e_due_process_5 >= e_due_process_14

    def test_unequal_enactment_text(self, e_search_clause, e_fourth_a):
        e_fourth_a.select_all()
        assert e_search_clause != e_fourth_a

    def test_enactment_subset(self, e_search_clause, e_fourth_a):
        e_fourth_a.select_all()
        assert e_search_clause < e_fourth_a

    def test_comparison_to_factor_false(self, e_due_process_5, watt_factor):
        f1 = watt_factor["f1"]
        assert not e_due_process_5 == f1

    def test_implication_of_factor_fails(self, e_due_process_5, watt_factor):
        f1 = watt_factor["f1"]
        with pytest.raises(TypeError):
            assert not e_due_process_5 > f1

    def test_implication_by_factor_fails(self, e_due_process_5, watt_factor):
        f1 = watt_factor["f1"]
        with pytest.raises(TypeError):
            assert not e_due_process_5 < f1

    def test_read_constitution_for_effective_date(self, fake_usc_client):
        ex_post_facto_provision = fake_usc_client.read_from_json(
            {"node": "/us/const/article/I/8/8"}
        )
        assert ex_post_facto_provision.start_date == datetime.date(1788, 9, 13)

    def test_bill_of_rights_effective_date(self, e_search_clause):
        # December 15, 1791
        assert e_search_clause.start_date == datetime.date(1791, 12, 15)

    def test_date_and_text_from_path_and_regime(self, fake_usc_client):
        """
        This tests assigning the full text of
        the section as the text of the Enactment, even though no
        ``exact``, ``prefix``, or ``suffix` parameter was
        passed to the TextQuoteSelector constructor.
        """
        amendment_12 = fake_usc_client.read_from_json(
            {"node": "/us/const/amendment/XIV"}
        )
        assert amendment_12.start_date == datetime.date(1868, 7, 28)
        assert "All persons born" in amendment_12.text

    def test_compare_effective_dates(self, e_due_process_5, e_due_process_14):
        assert e_due_process_14.start_date > e_due_process_5.start_date

    def test_invalid_selector_text(self, make_selector):
        with pytest.raises(TextSelectionError):
            _ = Enactment(
                node="/us/const/amendment/IV",
                selection=make_selector["bad_selector"],
                heading="",
                content="Not the same text as in the selector",
                start_date="2000-01-01",
            )

    # Addition

    def test_add_overlapping_enactments(self, e_search_clause, e_warrants_clause):
        combined = e_search_clause + e_warrants_clause

        passage = (
            "against unreasonable searches and seizures, "
            + "shall not be violated, "
            + "and no Warrants shall issue,"
        )
        assert passage in combined.text

    def test_add_shorter_plus_longer(self, e_fourth_a, e_search_clause):
        e_fourth_a.select_all()
        combined = e_search_clause + e_fourth_a

        assert combined.selected_text() == e_fourth_a.selected_text()
        assert combined.means(e_fourth_a)

    def test_add_longer_plus_shorter(self, e_fourth_a, e_search_clause):
        e_fourth_a.select_all()
        combined = e_fourth_a + e_search_clause

        assert combined.selected_text() == e_fourth_a.selected_text()
        assert combined.means(e_fourth_a)

    def test_consolidate_adjacent_passages(self, make_response):
        client = FakeClient(responses=make_response)
        copyright_clause = client.read("/us/const/article/I/8/8")
        copyright_statute = client.read("/us/usc/t17/s102/b")

        copyright_clause.select(None)
        securing_for_authors = copyright_clause + (
            "To promote the Progress of Science and "
            "useful Arts, by securing for limited Times to Authors"
        )
        and_inventors = copyright_clause + "and Inventors"
        right_to_writings = (
            copyright_clause + "the exclusive Right to their respective Writings"
        )
        to_combine = [
            copyright_statute,
            securing_for_authors,
            and_inventors,
            right_to_writings,
        ]
        combined = EnactmentGroup(to_combine)
        assert len(combined) == 2
        assert any(
            law.selected_text().startswith("To promote the Progress")
            and law.selected_text().endswith("their respective Writings…")
            for law in combined
        )

    def test_do_not_consolidate_from_different_sections(self, make_response):
        client = FakeClient(responses=make_response)

        due_process_5 = client.read("/us/const/amendment/V")
        due_process_14 = client.read("/us/const/amendment/XIV")

        due_process_5.select("life, liberty, or property, without due process of law")
        due_process_14.select("life, liberty, or property, without due process of law")

        combined = EnactmentGroup([due_process_5, due_process_14])
        assert len(combined) == 2

    def test_cannot_add_fact_to_enactment(self, watt_factor, e_search_clause):
        with pytest.raises(TypeError):
            print(e_search_clause + watt_factor["f3"])

    def test_cannot_add_enactment_to_statement(self, e_search_clause):
        statement = Statement(
            Predicate("$person committed a murder"), terms=Entity("Al")
        )
        with pytest.raises(TypeError):
            statement + e_search_clause

    def test_cannot_add_statement_to_enactment(self, e_search_clause):
        statement = Statement(
            Predicate("$person committed a murder"), terms=Entity("Al")
        )
        with pytest.raises(TypeError):
            e_search_clause + statement

    def test_cant_add_enactment_that_is_not_ancestor_or_descendant(
        self, e_search_clause, e_copyright
    ):
        with pytest.raises(ValueError):
            e_search_clause + e_copyright


class TestDump:
    client = Client(api_token=TOKEN)

    @pytest.mark.vcr
    def test_dump_dict(self):
        enactment = self.client.read("/us/const/article/I/8/8")
        d = dump.to_dict(enactment)
        start = d["selection"][0]["start"]
        end = d["selection"][0]["end"]
        text_selection = d["text_version"]["content"][start:end]
        assert "Science and useful Arts" in text_selection

    def test_dump_json(self, fake_beard_client):
        provision = fake_beard_client.read_from_json({"node": "/test/acts/47/6A"})
        dumped_provision = dump.to_json(provision)
        assert '"node": "/test/acts/47/6A"' in dumped_provision

    @pytest.mark.vcr
    def test_round_trip_dict(self, fake_beard_client):
        provision = fake_beard_client.read_from_json({"node": "/test/acts/47/6A"})
        dumped_provision = dump.to_dict(provision)
        new = self.client.read_from_json(dumped_provision)
        assert new.node == "/test/acts/47/6A"


class TestTextSelection:
    client = Client(api_token=TOKEN)

    def test_code_from_selector(self, fake_usc_client):
        enactment = fake_usc_client.read("/us/usc/t17/s103")
        assert enactment.code == "usc"

    def test_usc_selection(self, make_selector, fake_usc_client):
        enactment = fake_usc_client.read("/us/usc/t17/s103")
        selector = make_selector["preexisting material"]
        enactment.select(selector)

        assert enactment.level == CodeLevel.STATUTE
        assert enactment.jurisdiction == "us"
        assert enactment.code == "usc"

    def test_omit_terminal_slash(self, fake_usc_client):
        statute = fake_usc_client.read_from_json(
            {"exact": "process, system,", "node": "us/usc/t17/s102/b/"}
        )
        assert not statute.node.endswith("/")

    def test_add_omitted_initial_slash(self, fake_usc_client):
        statute = fake_usc_client.read_from_json(
            {"exact": "process, system,", "node": "us/usc/t17/s102/b/"}
        )
        assert statute.node.startswith("/")

    def test_selector_text_split(self):
        data = {"text": "process, system,|method of operation|, concept, principle"}
        schema = schemas_yaml.SelectorSchema()
        selector = schema.load(data)
        assert selector.exact.startswith("method")

    def test_whitespace_when_selecting_with_suffix(self, e_copyright):
        """Overwrite existing selector and test for trailing whitespace."""
        copyright_selector = TextQuoteSelector(suffix="idea, procedure,")
        e_copyright.select(copyright_selector)
        assert e_copyright.selected_text() == (
            "In no case does copyright protection "
            + "for an original work of authorship extend to any…"
        )

    def test_section_text_from_path(self, fake_usc_client):
        copyright_exceptions = fake_usc_client.read_from_json(
            {"node": "/us/usc/t17/s102/b"}
        )
        assert copyright_exceptions.text.startswith(
            "In no case does copyright protection "
            + "for an original work of authorship extend to any"
        )

    def test_exact_text_not_in_selection(self, fake_usc_client):
        with pytest.raises(TextSelectionError):
            _ = fake_usc_client.read_from_json(
                {
                    "node": "/us/const/amendment/XV/1",
                    "exact": "due process",
                }
            )

    def test_multiple_non_Factor_selectors_for_Holding(self):
        """
        The Holding-level TextQuoteSelectors should be built from this:

        "text": [
            "Census data therefore do not|trigger|copyright",
            "|may|possess the requisite originality"
            ]
        """
        client = FakeClient.from_file("usc.json")
        holdings, holding_anchors, _, _ = loaders.read_anchored_holdings_from_file(
            "holding_feist.json", client=client
        )
        assert len(holding_anchors[6]) == 2



"""Tests for any commands in readme.md."""

from authorityspoke.io.fake_enactments import FakeClient

from authorityspoke.io.loaders import (
    load_and_read_decision,
    read_anchored_holdings_from_file,
)
from authorityspoke.io.loaders import read_holdings_from_file


class TestReadme:
    def test_posit_holdings(self, make_response):
        client = FakeClient(responses=make_response)

        oracle = load_and_read_decision("oracle_h.json").majority
        lotus = load_and_read_decision("lotus_h.json").majority

        (
            oracle_holdings,
            oracle_anchors,
            oracle_named_anchors,
            oracle_e,
        ) = read_anchored_holdings_from_file("holding_oracle.json", client=client)
        (
            lotus_holdings,
            lotus_anchors,
            lotus_named_anchors,
            lotus_e,
        ) = read_anchored_holdings_from_file("holding_lotus.json", client=client)

        oracle.posit(
            holdings=oracle_holdings,
            holding_anchors=oracle_anchors,
            named_anchors=oracle_named_anchors,
            enactment_anchors=oracle_e,
        )
        lotus.posit(
            holdings=lotus_holdings,
            holding_anchors=lotus_anchors,
            named_anchors=lotus_named_anchors,
            enactment_anchors=lotus_e,
        )

        assert lotus.contradicts(oracle)

    def test_explain_contradiction(self, make_decision_with_holding):
        lotus = make_decision_with_holding["lotus"]
        oracle = make_decision_with_holding["oracle"]
        explanation = str(lotus.explain_contradiction(oracle))
        lotus_like = "<the Lotus menu command hierarchy> is like <the Java API>"
        java_like = "<the Java API> is like <the Lotus menu command hierarchy>"
        assert lotus_like in explanation or java_like in explanation



from copy import deepcopy
import datetime
import json
import os
import pathlib
from typing import Any, Dict, List, Text, Tuple

from anchorpoint.textselectors import TextQuoteSelector
from dotenv import load_dotenv
from legislice.download import Client
from legislice.yaml_schemas import ExpandableEnactmentSchema as EnactmentSchema
from nettlesome.terms import ContextRegister
from nettlesome.entities import Entity
from nettlesome.predicates import Predicate
from nettlesome.quantities import Comparison, Q_
import pytest

from authorityspoke.evidence import Evidence, Exhibit
from nettlesome.factors import Factor
from authorityspoke.facts import Fact, build_fact
from authorityspoke.holdings import Holding
from authorityspoke.opinions import Opinion
from authorityspoke.pleadings import Pleading, Allegation
from authorityspoke.rules import Procedure, Rule

from authorityspoke.io import loaders, readers
from authorityspoke.io.fake_enactments import FakeClient
from authorityspoke.io.schemas_json import RawFactor, RawHolding


load_dotenv()

TOKEN = os.getenv("LEGISLICE_API_TOKEN")
API_ROOT = "https://authorityspoke.com/api/v1"
legislice_client = Client(api_token=TOKEN)


@pytest.fixture(scope="module")
def vcr_config():
    return {
        # Replace the Authorization request header with "DUMMY" in cassettes
        "filter_headers": [("authorization", "DUMMY")],
    }


@pytest.fixture(scope="class")
def test_client() -> Client:
    client = Client(api_token=TOKEN, api_root=API_ROOT)
    client.coverage["/us/usc"] = {
        "latest_heading": "United States Code (USC)",
        "first_published": datetime.date(1926, 6, 30),
        "earliest_in_db": datetime.date(2013, 7, 18),
        "latest_in_db": datetime.date(2020, 8, 8),
    }
    client.coverage["/test/acts"] = {
        "latest_heading": "Test Acts",
        "first_published": datetime.date(1935, 4, 1),
        "earliest_in_db": datetime.date(1935, 4, 1),
        "latest_in_db": datetime.date(2013, 7, 18),
    }
    return client


@pytest.fixture(scope="class")
def section6d():
    return {
        "heading": "Waiver of beard tax in special circumstances",
        "content": "",
        "children": [
            {
                "heading": "",
                "content": "The Department of Beards shall waive the collection of beard tax upon issuance of beardcoin under Section 6C where the reason the maintainer wears a beard is due to bona fide religious, cultural, or medical reasons.",
                "children": [],
                "end_date": None,
                "node": "/test/acts/47/6D/1",
                "start_date": "2013-07-18",
                "url": "http://127.0.0.1:8000/api/v1/test/acts/47/6D/1@2018-03-11/",
            },
            {
                "heading": "",
                "content": "The determination of the Department of Beards as to what constitutes bona fide religious or cultural reasons shall be final and no right of appeal shall exist.",
                "children": [],
                "end_date": None,
                "node": "/test/acts/47/6D/2",
                "start_date": "1935-04-01",
                "url": "http://127.0.0.1:8000/api/v1/test/acts/47/6D/2@2018-03-11/",
            },
        ],
        "end_date": None,
        "node": "/test/acts/47/6D",
        "start_date": "1935-04-01",
        "url": "http://127.0.0.1:8000/api/v1/test/acts/47/6D@2018-03-11/",
        "parent": "http://127.0.0.1:8000/api/v1/test/acts/47@2018-03-11/",
    }


@pytest.fixture(scope="function")
def section_11_subdivided():
    return {
        "heading": "Licensed repurchasers of beardcoin",
        "start_date": "2013-07-18",
        "node": "/test/acts/47/11",
        "text_version": {
            "id": 1142710,
            "url": "https://authorityspoke.com/api/v1/textversions/1142710/",
            "content": "The Department of Beards may issue licenses to such",
        },
        "url": "https://authorityspoke.com/api/v1/test/acts/47/11/",
        "end_date": None,
        "children": [
            {
                "heading": "",
                "start_date": "2013-07-18",
                "node": "/test/acts/47/11/i",
                "text_version": {
                    "id": 1142704,
                    "url": "https://authorityspoke.com/api/v1/textversions/1142704/",
                    "content": "barbers,",
                },
                "url": "https://authorityspoke.com/api/v1/test/acts/47/11/i/",
                "end_date": None,
                "children": [],
                "citations": [],
            },
            {
                "heading": "",
                "start_date": "2013-07-18",
                "node": "/test/acts/47/11/ii",
                "text_version": {
                    "id": 1142705,
                    "url": "https://authorityspoke.com/api/v1/textversions/1142705/",
                    "content": "hairdressers, or",
                },
                "url": "https://authorityspoke.com/api/v1/test/acts/47/11/ii/",
                "end_date": None,
                "children": [],
                "citations": [],
            },
            {
                "heading": "",
                "start_date": "2013-07-18",
                "node": "/test/acts/47/11/iii",
                "text_version": {
                    "id": 1142706,
                    "url": "https://authorityspoke.com/api/v1/textversions/1142706/",
                    "content": "other male grooming professionals",
                },
                "url": "https://authorityspoke.com/api/v1/test/acts/47/11/iii/",
                "end_date": None,
                "children": [],
                "citations": [],
            },
            {
                "heading": "",
                "start_date": "2013-07-18",
                "node": "/test/acts/47/11/iii-con",
                "text_version": {
                    "id": 1142707,
                    "url": "https://authorityspoke.com/api/v1/textversions/1142707/",
                    "content": "as they see fit to purchase a beardcoin from a customer",
                },
                "url": "https://authorityspoke.com/api/v1/test/acts/47/11/iii-con/",
                "end_date": None,
                "children": [],
                "citations": [],
            },
            {
                "heading": "",
                "start_date": "2013-07-18",
                "node": "/test/acts/47/11/iv",
                "text_version": {
                    "id": 1142708,
                    "url": "https://authorityspoke.com/api/v1/textversions/1142708/",
                    "content": "whose beard they have removed,",
                },
                "url": "https://authorityspoke.com/api/v1/test/acts/47/11/iv/",
                "end_date": None,
                "children": [],
                "citations": [],
            },
            {
                "heading": "",
                "start_date": "2013-07-18",
                "node": "/test/acts/47/11/iv-con",
                "text_version": {
                    "id": 1142709,
                    "url": "https://authorityspoke.com/api/v1/textversions/1142709/",
                    "content": "and to resell those beardcoins to the Department of Beards.",
                },
                "url": "https://authorityspoke.com/api/v1/test/acts/47/11/iv-con/",
                "end_date": None,
                "children": [],
                "citations": [],
            },
        ],
        "citations": [],
        "parent": "https://authorityspoke.com/api/v1/test/acts/47/",
    }


@pytest.fixture(scope="module")
def fifth_a():
    return {
        "heading": "AMENDMENT V.",
        "content": "No person shall be held to answer for a capital, or otherwise infamous crime, unless on a presentment or indictment of a Grand Jury, except in cases arising in the land or naval forces, or in the Militia, when in actual service in time of War or public danger; nor shall any person be subject for the same offence to be twice put in jeopardy of life or limb; nor shall be compelled in any Criminal Case to be a witness against himself; nor be deprived of life, liberty, or property, without due process of law; nor shall private property be taken for public use, without just compensation.",
        "children": [],
        "end_date": None,
        "node": "/us/const/amendment/V",
        "start_date": "1791-12-15",
        "url": "https://authorityspoke.com/api/v1/us/const/amendment/V/",
        "parent": "https://authorityspoke.com/api/v1/us/const/amendment/",
        "selection": True,
    }


@pytest.fixture(scope="class")
def make_entity() -> Dict[str, Entity]:
    return {
        "motel": Entity("Hideaway Lodge"),
        "motel_specific": Entity("Hideaway Lodge", generic=False),
        "watt": Entity("Wattenburg"),
        "trees": Entity("the stockpile of trees"),
        "trees_specific": Entity("the stockpile of trees", generic=False),
        "tree_search": Entity("officers' search of the stockpile of trees"),
        "tree_search_specific": Entity(
            "officers' search of the stockpile of trees", generic=False
        ),
        "alice": Entity("Alice"),
        "bob": Entity("Bob"),
        "craig": Entity("Craig"),
        "dan": Entity("Dan"),
        "circus": Entity("circus"),
    }


@pytest.fixture(scope="class")
def make_predicate() -> Dict[str, Predicate]:

    return {
        "p1": Predicate("$place was a motel"),
        "p1_again": Predicate("$place was a motel"),
        "p2": Predicate("$person operated and lived at $place"),
        "p2_reflexive": Predicate("$person operated and lived at $person"),
        "p2_no_truth": Predicate("$person operated and lived at $place", truth=None),
        "p2_false": Predicate("$person operated and lived at $place", truth=False),
        "p3": Predicate("$place was ${person}’s abode"),
        "p3_false": Predicate("$place was ${person}’s abode", truth=False),
        "p4": Predicate("$thing was on the premises of $place"),
        "p5": Predicate("$thing was a stockpile of Christmas trees"),
        "p6": Predicate("$thing was among some standing trees"),
        "p7": Comparison(
            "the distance between $place1 and $place2 was",
            truth=False,
            sign=">",
            expression=Q_("35 feet"),
        ),
        "p7_obverse": Comparison(
            "the distance between $place1 and $place2 was",
            truth=True,
            sign="<=",
            expression=Q_("35 feet"),
        ),
        "p7_opposite": Comparison(
            "the distance between $place1 and $place2 was",
            truth=True,
            sign=">",
            expression=Q_("35 feet"),
        ),
        "p7_not_equal": Comparison(
            "the distance between $place1 and $place2 was",
            truth=True,
            sign="<>",
            expression=Q_("35 feet"),
        ),
        "p7_true": Comparison(
            "the distance between $place1 and $place2 was",
            truth=True,
            sign="<",
            expression=Q_("35 feet"),
        ),
        "p8": Comparison(
            "the distance between $place1 and $place2 was",
            sign=">=",
            expression=Q_("20 feet"),
        ),
        "p8_no_truth": Comparison(
            "the distance between $place1 and $place2 was",
            truth=None,
            sign=">=",
            expression=Q_("20 feet"),
        ),
        "p8_exact": Comparison(
            "the distance between $place1 and $place2 was",
            sign="==",
            expression=Q_("25 feet"),
        ),
        "p8_less": Comparison(
            "the distance between $place1 and $place2 was",
            sign="<=",
            expression=Q_("20 feet"),
        ),
        "p8_meters": Comparison(
            "the distance between $place1 and $place2 was",
            sign=">=",
            expression=Q_("10 meters"),
        ),
        "p8_int": Comparison(
            "the distance between $place1 and $place2 was",
            sign=">=",
            expression=20,
        ),
        "p8_float": Comparison(
            "the distance between $place1 and $place2 was",
            sign=">=",
            expression=20.0,
        ),
        "p8_higher_int": Comparison(
            "the distance between $place1 and $place2 was",
            sign=">=",
            expression=30,
        ),
        "p9": Comparison(
            "the distance between $thing and a parking area used by personnel and patrons of $place was",
            sign="<=",
            expression=Q_("5 feet"),
        ),
        "p9_exact": Comparison(
            "the distance between $thing and a parking area used by personnel and patrons of $place was",
            sign="=",
            expression=Q_("5 feet"),
        ),
        "p9_miles": Comparison(
            "the distance between $thing and a parking area used by personnel and patrons of $place was",
            sign="<=",
            expression=Q_("5 miles"),
        ),
        "p9_more": Comparison(
            "the distance between $thing and a parking area used by personnel and patrons of $place was",
            sign=">",
            expression=Q_("5 feet"),
        ),
        "p9_acres": Comparison(
            "the distance between $thing and a parking area used by personnel and patrons of $place was",
            sign="<=",
            expression=Q_("5 acres"),
        ),
        "p10": Predicate("$thing was within the curtilage of $place"),
        "p10_false": Predicate(
            "$thing was within the curtilage of $place", truth=False
        ),
        "p11": Predicate("$act was a warrantless search and seizure"),
        "p12": Predicate("$act was performed by federal law enforcement officers"),
        "p13": Predicate("$act constituted an intrusion upon $place"),
        "p14": Predicate("$person sought to preserve $thing as private"),
        "p15": Predicate("$thing was in an area adjacent to $place"),
        "p16": Predicate("$thing was in an area accessible to the public"),
        "p17": Predicate(
            "In $act, several law enforcement officials meticulously went through $thing"
        ),
        "p18": Comparison(
            "the length of time that $act continued was",
            sign=">=",
            expression=Q_("385 minutes"),
        ),
        "p19": Predicate("$act continued after night fell"),
        # Use the irrelevant predicates/factors to make sure they don't affect an outcome.
        "p_irrelevant_0": Predicate("$person was a clown"),
        "p_irrelevant_1": Predicate("$person was a bear"),
        "p_irrelevant_2": Predicate("$place was a circus"),
        "p_irrelevant_3": Predicate("$person performed at $place"),
        "p_crime": Predicate("$person committed a crime"),
        "p_murder": Predicate("$shooter murdered $victim"),
        "p_murder_whether": Predicate("$shooter murdered $victim", truth=None),
        "p_murder_false": Predicate("$shooter murdered $victim", truth=False),
        "p_irrelevant": Predicate("$evidence is relevant to show $fact", truth=False),
        "p_relevant": Predicate("$evidence is relevant to show $fact"),
        "p_relevant_whether": Predicate(
            "$evidence is relevant to show $fact", truth=None
        ),
        "p_shooting": Predicate("$shooter shot $victim"),
        "p_shooting_self": Predicate("$shooter shot $shooter"),
        "p_no_shooting": Predicate("$shooter shot $victim", truth=False),
        "p_shooting_whether": Predicate("$shooter shot $victim", truth=None),
        "p_no_crime": Predicate("$person1 committed a crime", truth=False),
        "p_three_entities": Predicate("$planner told $intermediary to hire $shooter"),
        "p_small_weight": Comparison(
            "the amount of gold $person possessed was",
            sign=">=",
            expression=Q_("1 gram"),
        ),
        "p_large_weight": Comparison(
            "the amount of gold $person possessed was",
            sign=">=",
            expression=Q_("100 kilograms"),
        ),
        "p_friends": Predicate("$person1 and $person2 were friends"),
        "p_reliable": Predicate("$evidence was reliable"),
        "p_quantity=3": Comparison("The number of mice was", sign="==", expression=3),
        "p_quantity>=4": Comparison("The number of mice was", sign=">=", expression=4),
        "p_quantity>5": Comparison("The number of mice was", sign=">", expression=5),
        "p_no_context": Predicate("context was included", truth=False),
    }


@pytest.fixture(scope="class")
def watt_mentioned(make_entity) -> Tuple[Entity, ...]:
    e = make_entity
    return (e["motel"], e["watt"], e["trees"], e["tree_search"], e["motel_specific"])


@pytest.fixture(scope="class")
def watt_factor(make_predicate, make_entity, watt_mentioned) -> Dict[str, Factor]:
    p = make_predicate

    c = watt_mentioned

    return {
        "f1": build_fact(p["p1"], case_factors=c),
        "f2": build_fact(p["p2"], (1, 0), case_factors=c),
        "f3": build_fact(p["p3"], case_factors=c),
        "f4": build_fact(p["p4"], (2, 0), case_factors=c),
        "f5": build_fact(p["p5"], 2, case_factors=c),
        "f6": build_fact(p["p6"], 2, case_factors=c),
        "f7": build_fact(p["p7"], (0, 2), case_factors=c),
        "f8": build_fact(p["p8"], (0, 2), case_factors=c),
        "f9": build_fact(p["p9"], (2, 0), case_factors=c),
        "f10": build_fact(p["p10"], (2, 0), case_factors=c),
        "f1_entity_order": build_fact(p["p1"], (1,), case_factors=c),
        "f1_different_entity": build_fact(p["p1"], (2,), case_factors=c),
        "f1_specific": build_fact(p["p1"], (4,), case_factors=c),
        "f1b": build_fact(p["p1"], case_factors=c),
        "f1c": build_fact(p["p1_again"], case_factors=c),
        "f2_preponderance_of_evidence": build_fact(
            p["p2"],
            (1, 0),
            standard_of_proof="preponderance of evidence",
            case_factors=c,
        ),
        "f2_clear_and_convincing": build_fact(
            p["p2"], (1, 0), standard_of_proof="clear and convincing", case_factors=c
        ),
        "f2_beyond_reasonable_doubt": build_fact(
            p["p2"], (1, 0), standard_of_proof="beyond reasonable doubt", case_factors=c
        ),
        "f2_different_entity": build_fact(p["p2"], (1, 2), case_factors=c),
        "f2_entity_order": build_fact(p["p2"], case_factors=c),
        "f2_no_truth": build_fact(p["p2_no_truth"], (1, 0), case_factors=c),
        "f2_false": build_fact(p["p2_false"], case_factors=c),
        "f2_false_absent": build_fact(p["p2_false"], absent=True, case_factors=c),
        "f2_reflexive": build_fact(p["p2_reflexive"], case_factors=c),
        "f2_generic": build_fact(p["p2"], generic=True, case_factors=c),
        "f2_false_generic": build_fact(p["p2_false"], generic=True, case_factors=c),
        "f3_generic": build_fact(p["p3"], generic=True, case_factors=c),
        "f3_different_entity": build_fact(p["p3"], (2, 1), case_factors=c),
        "f3_entity_order": build_fact(p["p3"], (1, 0), case_factors=c),
        "f3_absent": build_fact(p["p3"], absent=True, case_factors=c),
        "f4_h4": build_fact(p["p4"], (3, 0), case_factors=c),
        "f4_swap_entities": build_fact(p["p4"], (0, 2), case_factors=c),
        "f4_swap_entities_4": build_fact(p["p4"], (1, 4), case_factors=c),
        "f5_h4": build_fact(p["p5"], (3,), case_factors=c),
        "f5_swap_entities": build_fact(p["p5"], (0,), case_factors=c),
        "f6_swap_entities": build_fact(p["p6"], (0,), case_factors=c),
        "f7_opposite": build_fact(p["p7_opposite"], (0, 2), case_factors=c),
        "f7_swap_entities": build_fact(p["p7"], (2, 0), case_factors=c),
        "f7_swap_entities_4": build_fact(p["p7"], (1, 4), case_factors=c),
        "f7_true": build_fact(p["p7_true"], (0, 2), case_factors=c),
        "f8_absent": build_fact(p["p8"], (0, 2), absent=True, case_factors=c),
        "f8_exact": build_fact(p["p8_exact"], (0, 2), case_factors=c),
        "f8_exact_swap_entities": build_fact(p["p8_exact"], (2, 0), case_factors=c),
        "f8_float": build_fact(p["p8_float"], (0, 2), case_factors=c),
        "f8_higher_int": build_fact(p["p8_higher_int"], (0, 2), case_factors=c),
        "f8_int": build_fact(p["p8_int"], (0, 2), case_factors=c),
        "f8_less": build_fact(p["p8_less"], (0, 2), case_factors=c),
        "f8_less_absent": build_fact(p["p8_less"], (0, 2), absent=True, case_factors=c),
        "f8_meters": build_fact(p["p8_meters"], (0, 2), case_factors=c),
        "f8_no_truth": build_fact(p["p8_no_truth"], (0, 2), case_factors=c),
        "f9_absent": build_fact(p["p9"], (2, 0), absent=True, case_factors=c),
        "f9_miles": build_fact(p["p9_miles"], (2, 0), case_factors=c),
        "f9_absent_miles": build_fact(
            p["p9_miles"], (2, 0), absent=True, case_factors=c
        ),
        "f9_more_different_entity": Fact(
            p["p9_more"], (make_entity["circus"], make_entity["motel"])
        ),
        "f9_swap_entities": build_fact(p["p9"], (0, 2), case_factors=c),
        "f9_swap_entities_4": build_fact(p["p9"], (1, 4), case_factors=c),
        "f10_absent": build_fact(p["p10"], (2, 0), absent=True, case_factors=c),
        "f10_false": build_fact(p["p10_false"], (2, 0), case_factors=c),
        "f10_absent_false": build_fact(p["p10_false"], absent=True, case_factors=c),
        "f10_swap_entities": build_fact(p["p10"], (0, 2), case_factors=c),
        "f10_swap_entities_4": build_fact(p["p10"], (1, 4), case_factors=c),
        "f11": build_fact(p["p11"], 3, case_factors=c),
        "f12": build_fact(p["p12"], 3, case_factors=c),
        "f13": build_fact(p["p13"], (3, 2), case_factors=c),
        "f14": build_fact(p["p14"], (1, 2), case_factors=c),
        "f15": build_fact(p["p15"], (2, 0), case_factors=c),
        "f16": build_fact(p["p16"], 2, case_factors=c),
        "f17": build_fact(p["p17"], (2, 3), case_factors=c),
        "f18": build_fact(p["p18"], 2, case_factors=c),
        "f19": build_fact(p["p19"], 2, case_factors=c),
    }


@pytest.fixture(scope="class")
def make_factor(make_predicate, make_entity) -> Dict[str, Factor]:
    p = make_predicate
    e = make_entity

    c = (e["alice"], e["bob"], e["craig"], e["dan"], e["circus"])

    return {
        "f_irrelevant_0": build_fact(p["p_irrelevant_0"], (2,), case_factors=c),
        "f_irrelevant_1": build_fact(p["p_irrelevant_1"], (3,), case_factors=c),
        "f_irrelevant_2": build_fact(p["p_irrelevant_2"], (4,), case_factors=c),
        "f_irrelevant_3": build_fact(p["p_irrelevant_3"], (2, 4), case_factors=c),
        "f_irrelevant_3_new_context": build_fact(
            p["p_irrelevant_3"], (3, 4), case_factors=c
        ),
        "f_irrelevant_3_context_0": build_fact(
            p["p_irrelevant_3"], (3, 0), case_factors=c
        ),
        "f_crime": build_fact(p["p_crime"], case_factors=c),
        "f_crime_craig_poe": Fact(
            p["p_crime"],
            terms=e["craig"],
            standard_of_proof="preponderance of evidence",
        ),
        "f_watt_crime": build_fact(p["p_crime"], case_factors=make_entity["watt"]),
        "f_no_crime": build_fact(p["p_no_crime"], case_factors=c),
        "f_no_crime_entity_order": build_fact(p["p_no_crime"], (1,), case_factors=c),
        "f_murder": build_fact(p["p_murder"], case_factors=c),
        "f_no_murder": build_fact(p["p_murder_false"], case_factors=c),
        "f_murder_entity_swap": build_fact(p["p_murder"], (1, 0), case_factors=c),
        "f_murder_craig": build_fact(p["p_murder"], (2, 3), case_factors=c),
        "f_murder_whether": build_fact(p["p_murder_whether"], case_factors=c),
        "f_shooting": build_fact(p["p_shooting"], case_factors=c),
        "f_shooting_self": build_fact(p["p_shooting_self"], case_factors=c),
        "f_shooting_craig": build_fact(p["p_shooting"], (2, 3), case_factors=c),
        "f_shooting_craig_poe": build_fact(
            p["p_shooting"],
            (2, 3),
            case_factors=c,
            standard_of_proof="preponderance of evidence",
        ),
        "f_shooting_craig_brd": build_fact(
            p["p_shooting"],
            (2, 3),
            case_factors=c,
            standard_of_proof="beyond reasonable doubt",
        ),
        "f_shooting_entity_order": build_fact(p["p_shooting"], (1, 0), case_factors=c),
        "f_no_shooting": build_fact(p["p_no_shooting"], case_factors=c),
        "f_shooting_whether": build_fact(p["p_shooting_whether"], case_factors=c),
        "f_no_shooting_entity_order": build_fact(
            p["p_no_shooting"], (1, 0), case_factors=c
        ),
        "f_three_entities": build_fact(
            p["p_three_entities"], (0, 1, 2), case_factors=c
        ),
        "f_large_weight": build_fact(p["p_large_weight"], (0,), case_factors=c),
        "f_small_weight": build_fact(p["p_small_weight"], (0,), case_factors=c),
        "f_friends": build_fact(p["p_friends"], (0, 1), case_factors=c),
        "f_no_context": build_fact(p["p_no_context"], case_factors=c),
    }


@pytest.fixture(scope="class")
def make_exhibit(
    make_entity, make_predicate, make_factor, watt_factor, make_complex_fact
) -> Dict[str, Exhibit]:
    e = make_entity
    f = make_factor
    w = watt_factor
    c = make_complex_fact

    return {
        "shooting_affidavit": Exhibit(
            form="affidavit",
            statement=f["f_shooting"],
            statement_attribution=e["alice"],
        ),
        "shooting_testimony": Exhibit(
            form="testimony",
            statement=f["f_shooting"],
            statement_attribution=e["alice"],
        ),
        "no_shooting_testimony": Exhibit(
            form="testimony",
            statement=f["f_no_shooting"],
            statement_attribution=e["alice"],
        ),
        "no_shooting_entity_order_testimony": Exhibit(
            form="testimony",
            statement=f["f_no_shooting_entity_order"],
            statement_attribution=e["bob"],
        ),
        "no_shooting_witness_unknown_testimony": Exhibit(
            form="testimony", statement=f["f_no_shooting"]
        ),
        "no_shooting_witness_unknown_absent_testimony": Exhibit(
            form="testimony", statement=f["f_no_shooting"], absent=True
        ),
        "no_shooting_different_witness_testimony": Exhibit(
            form="testimony",
            statement=f["f_no_shooting"],
            statement_attribution=e["bob"],
        ),
        "reciprocal_testimony": Exhibit(
            form="testimony", statement=w["f8"], statement_attribution=e["craig"]
        ),
        "reciprocal_declaration": Exhibit(
            form="declaration", statement=w["f8"], statement_attribution=e["craig"]
        ),
        "reciprocal_testimony_absent": Exhibit(
            form="testimony",
            statement=w["f8"],
            statement_attribution=e["craig"],
            absent=True,
        ),
        "reciprocal_testimony_less": Exhibit(
            form="testimony", statement=w["f8_less"], statement_attribution=e["craig"]
        ),
        "reciprocal_testimony_specific": Exhibit(
            form="testimony", statement=w["f8_meters"], statement_attribution=e["craig"]
        ),
        "reciprocal_testimony_specific_absent": Exhibit(
            form="testimony",
            statement=w["f8_meters"],
            statement_attribution=e["craig"],
            absent=True,
        ),
        "relevant_murder_testimony": Exhibit(
            form="testimony",
            statement=c["f_relevant_murder"],
            statement_attribution=e["alice"],
        ),
        "relevant_murder_nested_swap_testimony": Exhibit(
            form="testimony",
            statement=c["f_relevant_murder_nested_swap"],
            statement_attribution=e["bob"],
        ),
        "relevant_murder_alice_craig_testimony": Exhibit(
            form="testimony",
            statement=c["f_relevant_murder_alice_craig"],
            statement_attribution=e["alice"],
        ),
        "large_weight_testimony": Exhibit(
            form="testimony",
            statement=f["f_large_weight"],
            statement_attribution=e["bob"],
        ),
        "small_weight_testimony": Exhibit(
            form="testimony",
            statement=f["f_small_weight"],
            statement_attribution=e["bob"],
        ),
        "generic_exhibit": Exhibit(generic=True),
        "specific_but_featureless": Exhibit(),
        "testimony_no_statement": Exhibit(form="testimony"),
    }


@pytest.fixture(scope="class")
def make_complex_fact(make_predicate, make_factor) -> Dict[str, Fact]:
    p = make_predicate
    f = make_factor

    return {
        "f_irrelevant_murder": Fact(
            p["p_irrelevant"], (f["f_shooting"], f["f_murder"])
        ),
        "f_relevant_murder": Fact(p["p_relevant"], (f["f_shooting"], f["f_murder"])),
        "f_relevant_murder_swap_entities": Fact(
            p["p_relevant"], (f["f_shooting"], f["f_murder"])
        ),
        "f_relevant_murder_nested_swap": Fact(
            p["p_relevant"], (f["f_shooting_entity_order"], f["f_murder_entity_swap"])
        ),
        "f_relevant_murder_whether": Fact(
            p["p_relevant"], (f["f_shooting"], f["f_murder_whether"])
        ),
        "f_whether_relevant_murder_whether": Fact(
            p["p_relevant"], (f["f_shooting_whether"], f["f_murder_whether"])
        ),
        "f_relevant_murder_swap": Fact(
            p["p_relevant"], (f["f_shooting"], f["f_murder_entity_swap"])
        ),
        "f_relevant_murder_craig": Fact(
            p["p_relevant"], (f["f_shooting_craig"], f["f_murder_craig"])
        ),
        "f_relevant_murder_alice_craig": Fact(
            p["p_relevant"], (f["f_shooting"], f["f_murder_craig"])
        ),
    }


@pytest.fixture(scope="class")
def make_fact_about_exhibit(make_predicate, make_exhibit) -> Dict[str, Evidence]:
    p = make_predicate
    e = make_exhibit
    return {
        "f_reliable_large_weight": Fact(
            p["p_reliable"], (e["large_weight_testimony"],)
        ),
        "f_reliable_small_weight": Fact(
            p["p_reliable"], (e["small_weight_testimony"],)
        ),
    }


@pytest.fixture(scope="class")
def make_complex_rule(
    make_factor, make_exhibit, make_complex_fact, make_fact_about_exhibit
) -> Dict[str, Rule]:
    return {
        "accept_relevance_testimony": Rule(
            Procedure(
                inputs=make_exhibit["relevant_murder_testimony"],
                outputs=make_complex_fact["f_relevant_murder"],
            )
        ),
        "accept_relevance_testimony_ALL": Rule(
            Procedure(
                inputs=make_exhibit["relevant_murder_testimony"],
                outputs=make_complex_fact["f_relevant_murder"],
            ),
            universal=True,
        ),
        "accept_murder_fact_from_relevance": Rule(
            Procedure(
                inputs=make_complex_fact["f_relevant_murder"],
                outputs=make_factor["f_murder"],
            )
        ),
        "accept_murder_fact_from_relevance_and_shooting": Rule(
            Procedure(
                inputs=[
                    make_complex_fact["f_relevant_murder"],
                    make_factor["f_shooting"],
                ],
                outputs=make_factor["f_murder"],
            )
        ),
        "accept_murder_fact_from_relevance_and_shooting_craig": Rule(
            Procedure(
                inputs=[
                    make_complex_fact["f_relevant_murder_craig"],
                    make_factor["f_shooting_craig"],
                ],
                outputs=make_factor["f_murder_craig"],
            )
        ),
        "accept_small_weight_reliable": Rule(
            Procedure(
                inputs=[make_factor["f_small_weight"], make_factor["f_friends"]],
                outputs=make_fact_about_exhibit["f_reliable_small_weight"],
            ),
            universal=True,
        ),
        "accept_small_weight_reliable_more_evidence": Rule(
            Procedure(
                inputs=[make_factor["f_large_weight"], make_factor["f_friends"]],
                outputs=make_fact_about_exhibit["f_reliable_small_weight"],
            ),
            universal=True,
        ),
        "accept_large_weight_reliable": Rule(
            Procedure(
                inputs=[make_factor["f_small_weight"], make_factor["f_friends"]],
                outputs=make_fact_about_exhibit["f_reliable_large_weight"],
            ),
            universal=True,
        ),
    }


@pytest.fixture(scope="class")
def make_evidence(
    make_predicate, make_factor, watt_factor, make_exhibit
) -> Dict[str, Evidence]:
    p = make_predicate
    f = make_factor
    w = watt_factor
    x = make_exhibit
    return {
        "shooting": Evidence(x["shooting_testimony"], to_effect=f["f_crime"]),
        "shooting_no_effect": Evidence(x["shooting_testimony"]),
        "no_shooting": Evidence(x["no_shooting_testimony"], to_effect=f["f_no_crime"]),
        "no_shooting_absent": Evidence(
            x["no_shooting_testimony"], to_effect=f["f_no_crime"], absent=True
        ),
        "no_shooting_entity_order": Evidence(
            x["no_shooting_entity_order_testimony"],
            to_effect=f["f_no_crime_entity_order"],
        ),
        "no_shooting_witness_unknown": Evidence(
            x["no_shooting_witness_unknown_testimony"], to_effect=f["f_no_crime"]
        ),
        "no_shooting_witness_unknown_absent": Evidence(
            x["no_shooting_witness_unknown_testimony"],
            to_effect=f["f_no_crime"],
            absent=True,
        ),
        # Here the Exhibit is absent, not the Evidence. Pointless distinction?
        "no_shooting_witness_unknown_absent_exhibit": Evidence(
            x["no_shooting_witness_unknown_absent_testimony"], to_effect=f["f_no_crime"]
        ),
        "no_shooting_no_effect_entity_order": Evidence(
            x["no_shooting_entity_order_testimony"]
        ),
        "no_shooting_different_witness": Evidence(
            x["no_shooting_different_witness_testimony"], to_effect=f["f_no_crime"]
        ),
        "reciprocal": Evidence(x["reciprocal_testimony"], to_effect=f["f_no_crime"]),
        "crime": Evidence(
            x["generic_exhibit"], to_effect=f["f_watt_crime"], generic=True
        ),
        "crime_absent": Evidence(
            x["generic_exhibit"], to_effect=f["f_watt_crime"], absent=True, generic=True
        ),
        "generic": Evidence(x["generic_exhibit"], generic=True),
        "generic_absent": Evidence(x["generic_exhibit"], absent=True, generic=True),
    }


@pytest.fixture(scope="class")
def make_pleading(make_entity) -> Dict[str, Dict[str, Pleading]]:
    return {"craig": Pleading(filer=make_entity["craig"])}


@pytest.fixture(scope="class")
def make_allegation(make_pleading, make_factor) -> Dict[str, Dict[str, Allegation]]:
    return {
        "shooting": Allegation(
            statement=make_factor["f_shooting"], pleading=make_pleading["craig"]
        )
    }


@pytest.fixture(scope="module")
def make_selector() -> Dict[str, TextQuoteSelector]:
    return {
        "bad_selector": TextQuoteSelector(exact="text that doesn't exist in the code"),
        "preexisting material": TextQuoteSelector(
            exact=(
                "protection for a work employing preexisting material in which "
                + "copyright subsists does not extend to any part of the work in "
                + "which such material has been used unlawfully."
            )
        ),
        "copyright": TextQuoteSelector(suffix="idea, procedure,"),
        "copyright_requires_originality": TextQuoteSelector(
            suffix="fixed in any tangible"
        ),
    }


@pytest.fixture(scope="module")
def make_response() -> Dict[str, Dict]:
    """Mock api responses"""
    this_directory = os.path.dirname(os.path.abspath(__file__))
    parent_directory = os.path.dirname(this_directory)
    responses_filepath = parent_directory + "/example_data/responses/usc.json"
    with open(responses_filepath, "r") as f:
        responses = json.load(f)
    return responses


@pytest.fixture(scope="module")
def fake_usc_client() -> FakeClient:
    return FakeClient.from_file("usc.json")


@pytest.fixture(scope="module")
def beard_response() -> Dict[str, Dict]:
    """Mock api responses"""
    this_directory = os.path.dirname(os.path.abspath(__file__))
    parent_directory = os.path.dirname(this_directory)
    responses_filepath = parent_directory + "/example_data/responses/beard_act.json"
    with open(responses_filepath, "r") as f:
        responses = json.load(f)
    return responses


@pytest.fixture(scope="module")
def fake_beard_client() -> FakeClient:
    return FakeClient.from_file("beard_act.json")


@pytest.fixture(scope="module")
def e_fourth_a(make_response):
    schema = EnactmentSchema()
    enactment = schema.load(make_response["/us/const/amendment/IV"]["1791-12-15"])
    enactment.select_all()
    return enactment


@pytest.fixture(scope="module")
def e_search_clause(make_response):
    schema = EnactmentSchema()
    enactment = schema.load(make_response["/us/const/amendment/IV"]["1791-12-15"])
    selector = TextQuoteSelector(suffix=", and no Warrants shall issue")
    enactment.select(selector)
    return enactment


@pytest.fixture(scope="module")
def e_warrants_clause(make_response):
    schema = EnactmentSchema()
    enactment = schema.load(make_response["/us/const/amendment/IV"]["1791-12-15"])
    enactment.select("shall not be violated, and no Warrants shall issue,")
    return enactment


@pytest.fixture(scope="module")
def e_due_process_5(make_response):
    schema = EnactmentSchema()
    enactment = schema.load(make_response["/us/const/amendment/V"]["1791-12-15"])
    enactment.select("life, liberty, or property, without due process of law")
    return enactment


@pytest.fixture(scope="module")
def e_due_process_14(make_response):
    schema = EnactmentSchema()
    fourteenth = make_response["/us/const/amendment/XIV"]["1868-07-28"]
    enactment = schema.load(fourteenth["children"][0])
    enactment.select("life, liberty, or property, without due process of law")
    return enactment


@pytest.fixture(scope="module")
def e_securing_for_authors(make_response):
    schema = EnactmentSchema()
    enactment = schema.load(make_response["/us/const/article/I/8/8"]["1788-09-13"])
    selector = TextQuoteSelector(
        exact=(
            "To promote the Progress of Science and "
            + "useful Arts, by securing for limited Times to Authors"
        )
    )
    enactment.select(selector)
    return enactment


@pytest.fixture(scope="module")
def e_securing_exclusive_right_to_writings(make_response):
    schema = EnactmentSchema()
    enactment = schema.load(make_response["/us/const/article/I/8/8"]["1788-09-13"])
    enactment.select(
        "To promote the Progress of Science and useful Arts, by securing for limited Times to Authors"
    )
    enactment.select_more("the exclusive Right to their respective Writings")
    return enactment


@pytest.fixture(scope="module")
def e_and_inventors(make_response):
    schema = EnactmentSchema()
    enactment = schema.load(make_response["/us/const/article/I/8/8"]["1788-09-13"])
    enactment.select("and Inventors")
    return enactment


@pytest.fixture(scope="module")
def e_right_to_writings(make_response):
    schema = EnactmentSchema()
    enactment = schema.load(make_response["/us/const/article/I/8/8"]["1788-09-13"])
    enactment.select("the exclusive Right to their respective Writings")
    return enactment


@pytest.fixture(scope="module")
def e_copyright_protection(make_response, make_selector):
    schema = EnactmentSchema()
    enactment = schema.load(make_response["/us/usc/t17/s102/a"]["2013-07-18"])
    enactment.select(TextQuoteSelector(suffix="Works of authorship include"))
    return enactment


@pytest.fixture(scope="module")
def e_copyright_requires_originality(make_response, make_selector):
    schema = EnactmentSchema()
    enactment = schema.load(make_response["/us/usc/t17/s102/a"]["2013-07-18"])
    enactment.select(make_selector["copyright_requires_originality"])
    return enactment


@pytest.fixture(scope="module")
def e_copyright(make_response, make_selector):
    schema = EnactmentSchema()
    enactment = schema.load(make_response["/us/usc/t17/s102/b"]["2013-07-18"])
    enactment.select(make_selector["copyright"])
    return enactment


@pytest.fixture(scope="module")
def e_copyright_exceptions(make_response):
    schema = EnactmentSchema()
    enactment = schema.load(make_response["/us/usc/t17/s102/b"]["2013-07-18"])
    enactment.select(TextQuoteSelector(suffix="idea, procedure, process"))
    return enactment


@pytest.fixture(scope="module")
def e_copyright_exceptions_full(make_response):
    schema = EnactmentSchema()
    enactment = schema.load(make_response["/us/usc/t17/s102/b"]["2013-07-18"])
    return enactment


@pytest.fixture(scope="module")
def e_in_no_case(make_response):
    schema = EnactmentSchema()
    enactment = schema.load(make_response["/us/usc/t17/s102/b"]["2013-07-18"])
    enactment.select(
        "In no case does copyright protection for an original work of authorship extend to any"
    )
    return enactment


@pytest.fixture(scope="module")
def e_method_of_operation(make_response):
    schema = EnactmentSchema()
    enactment = schema.load(make_response["/us/usc/t17/s102/b"]["2013-07-18"])
    enactment.select(TextQuoteSelector(exact="method of operation"))
    return enactment


@pytest.fixture(scope="module")
def e_compilation(make_response):
    schema = EnactmentSchema()
    enactment = schema.load(make_response["/us/usc/t17/s103/b"]["2013-07-18"])
    enactment.select("The copyright in a compilation")
    return enactment


@pytest.fixture(scope="module")
def e_preexisting_material(make_response):
    schema = EnactmentSchema()
    enactment = schema.load(make_response["/us/usc/t17/s103/b"]["2013-07-18"])
    enactment.select(
        "extends only to the material contributed by the author "
        "of such work, as distinguished from the preexisting "
        "material employed in the work, and does not imply any "
        "exclusive right in the preexisting material."
    )
    return enactment


@pytest.fixture(scope="class")
def make_procedure(make_evidence, make_factor, watt_factor) -> Dict[str, Procedure]:
    e = make_evidence
    f = watt_factor
    m = make_factor

    return {
        "c1": Procedure(outputs=(f["f3"],), inputs=(f["f1"], f["f2"])),
        "c2": Procedure(
            outputs=(f["f10"],),
            inputs=(f["f4"], f["f5"], f["f6"], f["f7"], f["f9"]),
            despite=(f["f8"],),
        ),
        "c3": Procedure(
            outputs=e["crime_absent"],
            inputs=(f["f3"], f["f11"], f["f12"], f["f13"], f["f14"], f["f15"]),
            despite=(f["f16"]),
        ),
        "c4": Procedure(
            outputs=f["f13"],
            inputs=(
                f["f1"],
                f["f2"],
                f["f4_h4"],
                f["f5_h4"],
                f["f11"],
                f["f12"],
                f["f17"],
                f["f18"],
                f["f19"],
            ),
        ),
        "c1_again": Procedure(outputs=(f["f3"],), inputs=(f["f1"], f["f2"])),
        "c1_entity_order": Procedure(
            outputs=(f["f3_entity_order"],),
            inputs=(f["f2_entity_order"], f["f1_entity_order"]),
        ),
        "c1_factor_and_entity_order": Procedure(
            outputs=(f["f3_entity_order"],),
            inputs=(f["f1_entity_order"], f["f2_entity_order"]),
        ),
        "c1_easy": Procedure(outputs=(f["f3"],), inputs=(f["f2"])),
        "c2_absent_despite": Procedure(
            outputs=(f["f10"],),
            inputs=(f["f4"], f["f5"], f["f6"], f["f7"]),
            despite=(f["f8_exact"], f["f9_absent"]),
        ),
        "c2_exact_quantity": Procedure(
            outputs=(f["f10"],),
            inputs=(f["f4"], f["f5"], f["f6"], f["f8_exact"], f["f9"]),
        ),
        "c2_higher_quantity": Procedure(
            outputs=(f["f10"],),
            inputs=(f["f4"], f["f5"], f["f6"], f["f7"], f["f8_meters"], f["f9"]),
        ),
        "c2_exact_in_despite": Procedure(
            outputs=(f["f10"],),
            inputs=(f["f4"], f["f5"], f["f6"], f["f7"], f["f9"]),
            despite=(f["f8_exact"],),
        ),
        "c2_exact_in_despite_entity_order": Procedure(
            outputs=(f["f10_swap_entities"],),
            inputs=(
                f["f4_swap_entities"],
                f["f5_swap_entities"],
                f["f6_swap_entities"],
                f["f7_swap_entities"],
                f["f9_swap_entities"],
            ),
            despite=(f["f8_exact_swap_entities"],),
        ),
        "c2_irrelevant_inputs": Procedure(
            outputs=(f["f10"],),
            inputs=(
                f["f4"],
                f["f5"],
                f["f6"],
                f["f7"],
                f["f9"],
                m["f_irrelevant_0"],
                m["f_irrelevant_1"],
                m["f_irrelevant_2"],
                m["f_irrelevant_3"],
                m["f_irrelevant_3_new_context"],
            ),
            despite=(f["f8"],),
        ),
        "c2_irrelevant_outputs": Procedure(
            outputs=(
                f["f10"],
                m["f_irrelevant_0"],
                m["f_irrelevant_1"],
                m["f_irrelevant_2"],
                m["f_irrelevant_3"],
                m["f_irrelevant_3_context_0"],
            ),
            inputs=(f["f4"], f["f5"], f["f6"], f["f7"], f["f9"]),
        ),
        "c2_irrelevant_despite": Procedure(
            outputs=(f["f10"],),
            inputs=(f["f4"], f["f5"], f["f6"], f["f7"], f["f9"]),
            despite=(
                f["f8"],
                m["f_irrelevant_0"],
                m["f_irrelevant_1"],
                m["f_irrelevant_2"],
                m["f_irrelevant_3"],
                m["f_irrelevant_3_new_context"],
            ),
        ),
        "c2_reciprocal_swap": Procedure(
            outputs=(f["f10"],),
            inputs=(f["f4"], f["f5"], f["f6"], f["f7_swap_entities"], f["f9"]),
            despite=(f["f8"],),
        ),
        "c2_nonreciprocal_swap": Procedure(
            outputs=(f["f10"],),
            inputs=(f["f4_swap_entities"], f["f5"], f["f6"], f["f7"], f["f9"]),
            despite=(f["f8"],),
        ),
        "c2_broad_output": Procedure(
            outputs=(f["f8_int"],), inputs=(f["f4"], f["f5"], f["f6"], f["f7"], f["f9"])
        ),
        "c2_narrow_output": Procedure(
            outputs=(f["f8_higher_int"],),
            inputs=(f["f4"], f["f5"], f["f6"], f["f7"], f["f9"]),
        ),
        "c2_output_absent": Procedure(
            outputs=(f["f10_absent"],),
            inputs=(f["f4"], f["f5"], f["f6"], f["f7"], f["f9"]),
            despite=(f["f8"],),
        ),
        "c2_output_false": Procedure(
            outputs=(f["f10_false"],),
            inputs=(f["f4"], f["f5"], f["f6"], f["f7"], f["f9"]),
            despite=(f["f8"],),
        ),
        "c2_output_absent_false": Procedure(
            outputs=(f["f10_absent_false"],),
            inputs=(f["f4"], f["f5"], f["f6"], f["f7"], f["f9"]),
            despite=(f["f8"],),
        ),
        "c_near_means_no_curtilage": Procedure(
            outputs=(f["f10_false"],), inputs=(f["f7_true"])
        ),
        "c_nearer_means_curtilage": Procedure(
            outputs=(f["f10"],), inputs=(f["f8_less"])
        ),
        "c_near_means_curtilage": Procedure(outputs=(f["f10"],), inputs=(f["f7"])),
        "c_near_means_curtilage_even_if": Procedure(
            outputs=(f["f10"],), inputs=(f["f7"]), despite=(f["f8"])
        ),
        "c_exact_means_no_curtilage": Procedure(
            outputs=(f["f10_false"],), inputs=(f["f8_exact"])
        ),
        "c_far_means_no_curtilage": Procedure(
            outputs=(f["f10_false"],), inputs=(f["f8"]), despite=(f["f7"])
        ),
        "c3_fewer_inputs": Procedure(
            outputs=e["crime_absent"],
            inputs=(f["f3"], f["f11"], f["f12"], f["f15"]),
            despite=(f["f16"]),
        ),
        "c_output_distance_less": Procedure(outputs=(f["f9"]), inputs=(f["f1"])),
        "c_output_farther_different_entity": Procedure(
            outputs=(f["f9_more_different_entity"]), inputs=(f["f1"])
        ),
        "c_no_context": Procedure(outputs=m["f_no_context"]),
    }


@pytest.fixture(scope="class")
def real_holding(make_procedure, e_search_clause) -> Dict[str, Rule]:
    """These holdings can be changed in case they don't accurately reflect
    what's in real cases, or in case there are API improvements that
    allow them to become more accurate. I'll try not to write any tests
    that depend on them remaining the same."""

    c = make_procedure

    return {
        "h1": Holding(rule=Rule(c["c1"], enactments=e_search_clause, mandatory=True)),
        "h2": Holding(rule=Rule(c["c2"], enactments=e_search_clause, mandatory=True)),
        "h3": Holding(rule=Rule(c["c3"], enactments=e_search_clause, mandatory=True)),
        "h4": Holding(rule=Rule(c["c4"], enactments=e_search_clause, mandatory=True)),
    }


@pytest.fixture(scope="class")
def make_rule(
    make_procedure, e_fourth_a, e_search_clause, e_due_process_5
) -> Dict[str, Rule]:
    c = make_procedure

    return {
        "h1": Rule(c["c1"], enactments=e_search_clause),
        "h2": Rule(c["c2"], enactments=e_search_clause),
        "h3": Rule(c["c3"], enactments=e_search_clause),
        "h1_again": Rule(c["c1"], enactments=e_search_clause),
        "h1_entity_order": Rule(c["c1_entity_order"], enactments=e_search_clause),
        "h1_easy": Rule(c["c1_easy"], enactments=e_search_clause),
        "h2_without_cite": Rule(c["c2"]),
        "h2_fourth_a_cite": Rule(c["c2"], enactments=e_fourth_a),
        "h2_despite_due_process": Rule(
            c["c2"],
            enactments=e_search_clause,
            enactments_despite=e_due_process_5,
        ),
        "h2_ALL_due_process": Rule(
            c["c2"],
            enactments=(e_search_clause, e_due_process_5),
            mandatory=False,
            universal=True,
        ),
        "h2_ALL": Rule(
            c["c2"], enactments=e_search_clause, mandatory=False, universal=True
        ),
        "h2_ALL_MAY_output_false": Rule(
            c["c2_output_false"],
            enactments=e_search_clause,
            mandatory=False,
            universal=True,
        ),
        "h2_ALL_MUST": Rule(
            c["c2"], enactments=e_search_clause, mandatory=True, universal=True
        ),
        "h2_ALL_MUST_output_false": Rule(
            c["c2_output_false"],
            enactments=e_search_clause,
            mandatory=True,
            universal=True,
        ),
        "h2_exact_quantity": Rule(c["c2_exact_quantity"], enactments=e_search_clause),
        "h2_irrelevant_inputs": Rule(
            c["c2_irrelevant_inputs"], enactments=e_search_clause
        ),
        "h2_irrelevant_inputs_ALL": Rule(
            c["c2_irrelevant_inputs"], enactments=e_search_clause, universal=True
        ),
        "h2_irrelevant_inputs_MUST": Rule(
            c["c2_irrelevant_inputs"], enactments=e_search_clause, mandatory=True
        ),
        "h2_irrelevant_inputs_ALL_MUST": Rule(
            c["c2_irrelevant_inputs"],
            enactments=e_search_clause,
            mandatory=True,
            universal=True,
        ),
        "h2_reciprocal_swap": Rule(c["c2_reciprocal_swap"], enactments=e_search_clause),
        "h2_exact_in_despite": Rule(
            c["c2_exact_in_despite"], enactments=e_search_clause
        ),
        "h2_exact_in_despite_ALL": Rule(
            c["c2_exact_in_despite"],
            enactments=e_search_clause,
            mandatory=False,
            universal=True,
        ),
        "h2_exact_in_despite_ALL_entity_order": Rule(
            c["c2_exact_in_despite_entity_order"],
            enactments=e_search_clause,
            mandatory=False,
            universal=True,
        ),
        "h2_exact_quantity_ALL": Rule(
            c["c2_exact_quantity"],
            enactments=e_search_clause,
            mandatory=False,
            universal=True,
        ),
        "h2_MUST": Rule(
            c["c2"], enactments=e_search_clause, mandatory=True, universal=False
        ),
        "h2_output_absent": Rule(c["c2_output_absent"], enactments=e_search_clause),
        "h2_output_false": Rule(c["c2_output_false"], enactments=e_search_clause),
        "h2_output_false_ALL": Rule(
            c["c2_output_false"], enactments=e_search_clause, universal=True
        ),
        "h2_output_false_ALL_MUST": Rule(
            c["c2_output_false"],
            enactments=e_search_clause,
            mandatory=True,
            universal=True,
        ),
        "h2_output_absent_false": Rule(
            c["c2_output_absent_false"], enactments=e_search_clause
        ),
        "h2_SOME_MUST_output_false": Rule(
            c["c2_output_false"],
            enactments=e_search_clause,
            mandatory=True,
            universal=False,
        ),
        "h2_SOME_MUST_output_absent": Rule(
            c["c2_output_absent"],
            enactments=e_search_clause,
            mandatory=True,
            universal=False,
        ),
        "h3_ALL": Rule(c["c3"], enactments=e_search_clause, universal=True),
        "h3_fewer_inputs": Rule(c["c3_fewer_inputs"], enactments=e_search_clause),
        "h3_fewer_inputs_ALL": Rule(
            c["c3_fewer_inputs"], enactments=e_search_clause, universal=True
        ),
        "h_near_means_curtilage": Rule(
            c["c_near_means_curtilage"], enactments=e_search_clause
        ),
        "h_near_means_curtilage_even_if": Rule(
            c["c_near_means_curtilage_even_if"], enactments=e_search_clause
        ),
        "h_near_means_curtilage_ALL": Rule(
            c["c_near_means_curtilage"], enactments=e_search_clause, universal=True
        ),
        "h_near_means_curtilage_ALL_MUST": Rule(
            c["c_near_means_curtilage"],
            enactments=e_search_clause,
            mandatory=True,
            universal=True,
        ),
        "h_near_means_no_curtilage": Rule(
            c["c_near_means_no_curtilage"], enactments=e_search_clause
        ),
        "h_near_means_no_curtilage_ALL": Rule(
            c["c_near_means_no_curtilage"],
            enactments=e_search_clause,
            universal=True,
        ),
        "h_near_means_no_curtilage_ALL_MUST": Rule(
            c["c_near_means_no_curtilage"],
            enactments=e_search_clause,
            mandatory=True,
            universal=True,
        ),
        "h_nearer_means_curtilage": Rule(
            c["c_nearer_means_curtilage"], enactments=e_search_clause
        ),
        "h_nearer_means_curtilage_ALL": Rule(
            c["c_nearer_means_curtilage"], enactments=e_search_clause, universal=True
        ),
        "h_nearer_means_curtilage_MUST": Rule(
            c["c_nearer_means_curtilage"], enactments=e_search_clause, mandatory=True
        ),
        "h_far_means_no_curtilage": Rule(c["c_far_means_no_curtilage"]),
        "h_exact_means_no_curtilage_ALL": Rule(
            c["c_exact_means_no_curtilage"], universal=True
        ),
        "h_far_means_no_curtilage_ALL": Rule(
            c["c_far_means_no_curtilage"], universal=True
        ),
        "h_output_distance_less": Rule(
            c["c_output_distance_less"], universal=True, mandatory=True
        ),
        "h_output_farther_different_entity": Rule(
            c["c_output_farther_different_entity"]
        ),
    }


@pytest.fixture(scope="class")
def make_beard_rule(beard_response) -> List[Rule]:
    """Rules from the "Beard Tax Act" example statutes."""
    client = FakeClient(responses=beard_response)
    beard_dictionary = loaders.load_holdings("beard_rules.json")
    return readers.read_rules(beard_dictionary, client=client)


@pytest.fixture(scope="class")
def make_holding(make_rule) -> Dict[str, Holding]:
    holdings: Dict[str, Holding] = {}
    for name, rule in make_rule.items():
        holdings[name] = Holding(rule=rule)

    new_holdings = {
        "h1_opposite": Holding(rule=make_rule["h1"], rule_valid=False),
        "h2_undecided": Holding(rule=make_rule["h2"], decided=False),
        "h2_invalid_undecided": Holding(
            rule=make_rule["h2"], rule_valid=False, decided=False
        ),
        "h2_MUST_undecided": Holding(rule=make_rule["h2_MUST"], decided=False),
        "h2_MUST_invalid": Holding(rule=make_rule["h2_MUST"], rule_valid=False),
        "h2_invalid": Holding(rule=make_rule["h2"], rule_valid=False),
        "h2_ALL_invalid": Holding(rule=make_rule["h2_ALL"], rule_valid=False),
        "h2_irrelevant_inputs_undecided": Holding(
            rule=make_rule["h2_irrelevant_inputs"], decided=False
        ),
        "h2_irrelevant_inputs_invalid": Holding(
            rule=make_rule["h2_irrelevant_inputs"], rule_valid=False
        ),
        "h2_irrelevant_inputs_ALL_invalid": Holding(
            rule=make_rule["h2_irrelevant_inputs_ALL"], rule_valid=False
        ),
        "h2_irrelevant_inputs_ALL_MUST_invalid": Holding(
            rule=make_rule["h2_irrelevant_inputs_ALL_MUST"], rule_valid=False
        ),
        "h2_irrelevant_inputs_MUST_invalid": Holding(
            rule=make_rule["h2_irrelevant_inputs_MUST"], rule_valid=False
        ),
        "h2_ALL_due_process_invalid": Holding(
            rule=make_rule["h2_ALL_due_process"], rule_valid=False
        ),
        "h3_undecided": Holding(rule=make_rule["h3"], decided=False),
        "h3_ALL_undecided": Holding(rule=make_rule["h3_ALL"], decided=False),
        "h3_fewer_inputs_undecided": Holding(
            rule=make_rule["h3_fewer_inputs"], decided=False
        ),
        "h3_fewer_inputs_ALL_undecided": Holding(
            rule=make_rule["h3_fewer_inputs_ALL"], decided=False
        ),
        "h_near_means_curtilage_ALL_undecided": Holding(
            rule=make_rule["h_near_means_curtilage_ALL"], decided=False
        ),
    }

    holdings.update(new_holdings)
    return holdings


TEST_CASES = ("feist", "lotus", "oracle", "brad", "cardenas", "watt")


def load_decisions_for_fixtures():
    decisions = {}
    for case in TEST_CASES:
        decision = loaders.load_decision(f"{case}_h.json")
        built = readers.read_decision(decision)
        decisions[case] = built
    return decisions


@pytest.fixture(scope="class")
def make_decision():
    return load_decisions_for_fixtures()


@pytest.fixture(scope="class")
def make_decision_with_holding(make_response):
    client_without_api_access = FakeClient(responses=make_response)
    decisions = load_decisions_for_fixtures()
    for case in TEST_CASES:
        (
            holdings,
            holding_anchors,
            named_anchors,
            enactment_anchors,
        ) = loaders.read_anchored_holdings_from_file(
            f"holding_{case}.json",
            client=client_without_api_access,
        )
        decisions[case].majority.posit(
            holdings,
            holding_anchors=holding_anchors,
            named_anchors=named_anchors,
            enactment_anchors=enactment_anchors,
        )
    return decisions


@pytest.fixture(scope="class")
def make_opinion(make_decision) -> Dict[str, Opinion]:
    opinions = {}
    for case in TEST_CASES:
        for opinion in make_decision[case].opinions:
            opinions[f"{case}_{opinion.position}"] = opinion
    return opinions


@pytest.fixture(scope="class")
def make_opinion_with_holding(make_decision_with_holding) -> Dict[str, Opinion]:
    opinions = {}
    for case in TEST_CASES:
        for opinion in make_decision_with_holding[case].opinions:
            opinions[f"{case}_{opinion.position}"] = opinion
    return opinions


@pytest.fixture(scope="class")
def make_analysis() -> Dict[str, Dict[str, Any]]:
    """Example user analysis data."""
    return {
        "minimal": {
            "holdings": [
                {
                    "anchors": "upholding searches in |open fields or grounds|around a house",
                    "outputs": {
                        "type": "fact",
                        "content": "{Bradley} made a minimal holding object",
                    },
                }
            ],
        },
        "no anchors": {
            "holdings": [
                {
                    "outputs": {
                        "type": "fact",
                        "content": "this holding has no text anchors",
                    }
                }
            ]
        },
    }


@pytest.fixture(scope="function")
def raw_factor() -> RawFactor:
    """
    Example user data that needs to be processed and loaded as a Factor.
    """
    return {
        "relevant": {
            "content": "{} is relevant to show {}",
            "type": "Fact",
            "name": "relevant fact",
            "terms": [
                {"content": "{Short Name} shot {Longer Name}", "type": "Fact"},
                {
                    "content": "{} murdered {}",
                    "terms": ["Short Name", "Longer Name"],
                    "type": "Fact",
                },
            ],
        }
    }


@pytest.fixture(scope="function")
def raw_holding() -> RawHolding:
    return {
        "bradley_house": {
            "inputs": {"type": "fact", "content": "{Bradley} lived at Bradley's house"},
            "outputs": {
                "type": "evidence",
                "to_effect": {
                    "type": "fact",
                    "name": "fact that Bradley committed a crime",
                    "content": "Bradley committed a crime",
                },
                "name": "evidence of Bradley's guilt",
                "absent": True,
            },
        },
        "stolen watch": {
            "anchors": [
                {"exact": "Mark stole a watch"},
                {"exact": "a watch was stolen by Mark"},
            ],
            "outputs": [
                {"type": "Fact", "content": "{Mark} stole a watch"},
            ],
            "inputs": [
                {
                    "type": "Evidence",
                    "to_effect": {"type": "Fact", "content": "{Mark} stole a watch"},
                }
            ],
        },
    }


@pytest.fixture(scope="function")
def make_context_register() -> ContextRegister:
    context_names = ContextRegister()
    context_names.insert_pair(key=Entity("Alice"), value=Entity("Craig"))
    context_names.insert_pair(key=Entity("Bob"), value=Entity("Dan"))
    return context_names



"""Functions for saving objects to file after they have been JSON serialized."""
import json
import pathlib

from typing import Dict, List, Optional

from authorityspoke.io import filepaths


def case_to_file(
    case: Dict,
    filename: Optional[str] = None,
    directory: Optional[pathlib.Path] = None,
    filepath: Optional[pathlib.Path] = None,
) -> None:
    r"""
    Save one case from an API response as a JSON file.

    :param results:
        A dict representing a case, in the format
        of the Caselaw Access Project API.

    :param filename:
        Filename (not including the directory) for the
        file where the downloaded opinion should be saved.

    :param directory:
        A :py:class:`~pathlib.Path` object specifying the directory where the
        downloaded opinion should be saved. If ``None`` is given, the current
        default is ``example_data/cases``\.

    :param filepath:
        Complete path to the location where the JSON file should be saved,
        including filename.
    """
    validated_filepath = filepaths.make_filepath(
        filename, directory, filepath, default_folder="cases"
    )
    with open(validated_filepath, "w") as fp:
        json.dump(case, fp, ensure_ascii=False, indent=4)


def cases_to_file(
    results: List[Dict],
    filename: Optional[str] = None,
    directory: Optional[pathlib.Path] = None,
    filepath: Optional[pathlib.Path] = None,
) -> None:
    """
    Save cases from an API response as JSON files.

    :param results:
        A list of dicts representing cases, in the format
        of the Caselaw Access Project API.

    :param filename:
        Filename (not including the directory) for the
        file where the downloaded opinion should be saved.

    :param directory:
        A :py:class:`~pathlib.Path` object specifying the directory where the
        downloaded opinion should be saved. If ``None`` is given, the current
        default is ``example_data/cases``.

    :param filepath:
        Complete path to the location where the JSON file should be saved,
        including filename.
    """
    for number, case in enumerate(results):
        if not filename:
            mangled_filename = f'{case["id"]}.json'
        else:
            mangled_filename = filename
            if number > 0:
                mangled_filename = mangled_filename.replace(".", f"_{number}.")
        case_to_file(
            case=case, filename=mangled_filename, directory=directory, filepath=filepath
        )
    return None



from dataclasses import dataclass

from typing import ClassVar, Optional, Tuple

from nettlesome.entities import Entity
from nettlesome.formatting import indented

from nettlesome.factors import Factor
from authorityspoke.facts import Fact


@dataclass()
class Pleading(Factor):
    r"""
    A document filed by a party to make :class:`Allegation`\s.

    :param filer:
        the :class:`.Entity` that the pleading references as having filed it,
        regardless of any questions about the identification of the filer.

    :param name:

    :param absent:

    :param generic:
    """

    filer: Optional[Entity] = None
    name: Optional[str] = None
    absent: bool = False
    generic: bool = False
    context_factor_names: ClassVar[Tuple[str]] = ("filer",)

    def __str__(self):
        string = f'{("filed by " + self.filer.short_string if self.filer else "")}'
        return super().__str__().format(string)

    @property
    def short_string(self):
        return str(self)


@dataclass()
class Allegation(Factor):
    """
    A formal assertion of a :class:`Fact`.

    May be included by a party in a :class:`Pleading`
    to establish a cause of action.

    :param statement:
        a :class:`Fact` being alleged

    :param pleading:
        the :class:`Pleading` in where the allegation appears

    :param name:

    :param absent:

    :param generic:
    """

    statement: Optional[Fact] = None
    pleading: Optional[Pleading] = None
    name: Optional[str] = None
    absent: bool = False
    generic: bool = False
    context_factor_names: ClassVar[Tuple[str, ...]] = ("statement", "pleading")

    @property
    def wrapped_string(self):
        text = ""
        if self.statement:
            text += "\n" + indented("OF:")
            factor_text = indented(self.statement.wrapped_string, tabs=2)
            text += f"\n{str(factor_text)}"
        if self.pleading:
            text += f"\n" + indented("FOUND IN:")
            factor_text = indented(str(self.pleading), tabs=2)
            text += f"\n{str(factor_text)}"
        return super().__str__().format(text).strip()

    def __str__(self):
        string = (
            f'{("in " + self.pleading.short_string + ",") if self.pleading else ""}'
            + f'{("claiming " + self.statement.short_string + ",") if self.statement else ""}'
        )
        string = string.strip(",")
        return super().__str__().format(string).replace("Allegation", "allegation")



"""Downloading data that can be converted to authorityspoke objects."""

from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

import eyecite
from eyecite.models import CaseCitation
from legislice import Client as LegisClient
import requests

from authorityspoke.decisions import CAPCitation, Decision
from authorityspoke.io.schemas_json import RawDecision, DecisionSchema


class AuthoritySpokeAPIError(Exception):
    """Error for invalid API query."""

    pass


def normalize_case_cite(cite: Union[str, CaseCitation, CAPCitation]) -> str:
    """Get just the text that identifies a citation."""
    if isinstance(cite, CAPCitation):
        return cite.cite
    if isinstance(cite, str):
        possible_cites = list(eyecite.get_citations(cite))
        bad_cites = []
        for possible in possible_cites:
            if isinstance(possible, CaseCitation):
                return possible.corrected_citation()
            bad_cites.append(possible)
        error_msg = f"Could not locate a CaseCitation in the text {cite}."
        for bad_cite in bad_cites:
            error_msg += f" {str(bad_cite)} was type {bad_cite.__class__.__name__}, not CaseCitation."
        raise ValueError(error_msg)
    return cite.corrected_citation()


class CAPClient:
    """Downloads Decisions from Case Access Project API."""

    def __init__(self, api_token: str = ""):

        """Create download client with an API token and an API address."""
        self.endpoint = f"https://api.case.law/v1/cases/"
        if api_token and api_token.startswith("Token "):
            api_token = api_token.split("Token ")[1]
        self.api_token = api_token or ""

    def get_api_headers(self, full_case=bool) -> Dict[str, str]:
        """Get API headers based on whether the full case text is requested."""
        api_dict = {}
        if full_case:
            if not self.api_token:
                raise AuthoritySpokeAPIError(
                    "To fetch full opinion text using the full_case parameter, "
                    "set the CAPClient's 'api_key' attribute to "
                    "your API key for the Case Access Project. See https://api.case.law/"
                )
            api_dict["Authorization"] = f"Token {self.api_token}"
        return api_dict

    def fetch_decision_list_by_cite(
        self, cite: Union[str, CaseCitation, CAPCitation], full_case: bool = False
    ) -> List[RawDecision]:
        """
        Get the "results" list for a queried citation from the CAP API.

        :param cite:
            a citation linked to an opinion in the
            `Caselaw Access Project database <https://case.law/api/>`_.
            Usually these will be in the traditional format
            ``[Volume Number] [Reporter Name Abbreviation] [Page Number]``, e.g.
            `750 F.3d 1339 <https://case.law/search/#/cases?page=1&cite=%22750%20F.3d%201339%22>`_
            for Oracle America, Inc. v. Google Inc.

        :param full_case:
            whether to request the full text of the opinion from the
            `Caselaw Access Project API <https://api.case.law/v1/cases/>`_.
            If this is ``True``, the CAPClient must have the `api_token` attribute.

        :returns:
            the "results" list for this queried citation.
        """

        normalized_cite = normalize_case_cite(cite)

        params = {"cite": normalized_cite}

        headers = self.get_api_headers(full_case=full_case)

        if full_case:
            params["full_case"] = "true"
        response = requests.get(self.endpoint, params=params, headers=headers).json()
        return response["results"]

    def read_decision_list_by_cite(
        self, cite: Union[str, CaseCitation, CAPCitation], full_case: bool = False
    ) -> Decision:
        """
        Download and deserialize the "results" list for a queried citation from the CAP API.

        :param cite:
            a citation linked to an opinion in the
            `Caselaw Access Project database <https://case.law/api/>`_.
            Usually these will be in the traditional format
            ``[Volume Number] [Reporter Name Abbreviation] [Page Number]``, e.g.
            `750 F.3d 1339 <https://case.law/search/#/cases?page=1&cite=%22750%20F.3d%201339%22>`_
            for Oracle America, Inc. v. Google Inc.

        :param full_case:
            whether to request the full text of the opinion from the
            `Caselaw Access Project API <https://api.case.law/v1/cases/>`_.
            If this is ``True``, the CAPClient must have the `api_token` attribute.

        :returns:
            the first case in the "results" list for this queried citation.
        """
        response = self.fetch_decision_list_by_cite(cite=cite, full_case=full_case)
        schema = DecisionSchema(many=True)
        return schema.load(response)

    def fetch_cite(
        self, cite: Union[str, CaseCitation, CAPCitation], full_case: bool = False
    ) -> RawDecision:
        """
        Download a decision from Caselaw Access Project API.

        :param cite:
            a citation linked to an opinion in the
            `Caselaw Access Project database <https://case.law/api/>`_.
            Usually these will be in the traditional format
            ``[Volume Number] [Reporter Name Abbreviation] [Page Number]``, e.g.
            `750 F.3d 1339 <https://case.law/search/#/cases?page=1&cite=%22750%20F.3d%201339%22>`_
            for Oracle America, Inc. v. Google Inc.

        :param full_case:
            whether to request the full text of the opinion from the
            `Caselaw Access Project API <https://api.case.law/v1/cases/>`_.
            If this is ``True``, the CAPClient must have the `api_token` attribute.

        :returns:
            the first case in the "results" list for this queried citation.
        """
        result_list = self.fetch_decision_list_by_cite(cite=cite, full_case=full_case)
        return result_list[0]

    def read_cite(
        self, cite: Union[str, CaseCitation, CAPCitation], full_case: bool = False
    ) -> Decision:
        """
        Download and deserialize a Decision from Caselaw Access Project API.

        :param cite:
            a citation linked to an opinion in the
            `Caselaw Access Project database <https://case.law/api/>`_.
            Usually these will be in the traditional format
            ``[Volume Number] [Reporter Name Abbreviation] [Page Number]``, e.g.
            `750 F.3d 1339 <https://case.law/search/#/cases?page=1&cite=%22750%20F.3d%201339%22>`_
            for Oracle America, Inc. v. Google Inc.

        :param full_case:
            whether to request the full text of the opinion from the
            `Caselaw Access Project API <https://api.case.law/v1/cases/>`_.
            If this is ``True``, the CAPClient must have the `api_token` attribute.

        :returns:
            the first case in the "results" list for this queried citation.
        """
        response = self.fetch_cite(cite=cite, full_case=full_case)
        schema = DecisionSchema()
        return schema.load(response)

    def fetch_id(self, cap_id: int, full_case: bool = False) -> RawDecision:
        """
        Download a decision from Caselaw Access Project API.

        :param cap_id:
            an identifier for an opinion in the
            `Caselaw Access Project database <https://case.law/api/>`_,
            e.g. 4066790 for
            `Oracle America, Inc. v. Google Inc. <https://api.case.law/v1/cases/4066790/>`_.

        :param full_case:
            whether to request the full text of the opinion from the
            `Caselaw Access Project API <https://api.case.law/v1/cases/>`_.
            If this is ``True``, the CAPClient must have the `api_token` attribute.

        :returns:
            the first case in the "results" list for this queried citation.
        """
        url = self.endpoint + f"{cap_id}/"
        headers = self.get_api_headers(full_case=full_case)
        params = {}
        if full_case:
            params["full_case"] = "true"
        response = requests.get(url, params=params, headers=headers).json()
        if cap_id and response.get("detail") == "Not found.":
            raise AuthoritySpokeAPIError(f"API returned no cases with id {cap_id}")
        return response

    def read_id(self, cap_id: int, full_case: bool = False) -> Decision:
        """
        Download a decision from Caselaw Access Project API.

        :param cap_id:
            an identifier for an opinion in the
            `Caselaw Access Project database <https://case.law/api/>`_,
            e.g. 4066790 for
            `Oracle America, Inc. v. Google Inc. <https://api.case.law/v1/cases/4066790/>`_.

        :param full_case:
            whether to request the full text of the opinion from the
            `Caselaw Access Project API <https://api.case.law/v1/cases/>`_.
            If this is ``True``, the CAPClient must have the `api_token` attribute.

        :returns:
            a Decision created from the first case in the "results" list for
            this queried citation.
        """
        response = self.fetch_id(cap_id=cap_id, full_case=full_case)
        schema = DecisionSchema()
        return schema.load(response)

    def fetch(
        self, query: Union[int, str, CaseCitation, CAPCitation], full_case: bool = False
    ) -> RawDecision:
        """Query by CAP id or citation, and download Decision from CAP API."""
        if isinstance(query, int) or (isinstance(query, str) and query.isdigit()):
            return self.fetch_id(int(query), full_case=full_case)
        return self.fetch_cite(query)

    def read(
        self, query: Union[int, str, CaseCitation, CAPCitation], full_case: bool = False
    ) -> Decision:
        """Query by CAP id or citation, download, and load Decision from CAP API."""
        if isinstance(query, int) or (isinstance(query, str) and query.isdigit()):
            return self.read_id(int(query), full_case=full_case)
        return self.read_cite(query)



import logging

from nettlesome import Entity

from authorityspoke.evidence import Exhibit
from authorityspoke.facts import Fact, Predicate


class TestExhibits:
    def test_make_exhibit_object(self):
        e = Exhibit(form="testimony")
        assert not e.absent

    def test_exhibit_short_string(self, make_exhibit):
        assert (
            make_exhibit["no_shooting_testimony"].short_string.lower()
            == (
                "the testimony attributed to <Alice>, asserting "
                "the fact it was false that <Alice> shot <Bob>,"
            ).lower()
        )

    def test_comma_when_exhibit_is_in_fact(self):
        coin = Exhibit(
            form="token",
            statement=Fact(
                predicate=Predicate(
                    "$agency granted an exemption from the prohibition of wearing beards"
                ),
                terms=Entity("the Department of Beards"),
            ),
            statement_attribution=Entity("the Department of Beards"),
        )
        counterfeit = Fact(predicate="$thing was counterfeit", terms=coin)
        assert str(counterfeit) == (
            "the fact that the token attributed to <the Department of Beards>, "
            "asserting the fact that <the Department of Beards> granted an "
            "exemption from the prohibition of wearing beards, was counterfeit"
        )


class TestExhibitsSameMeaning:
    def test_equality(self, make_exhibit):
        assert make_exhibit["no_shooting_entity_order_testimony"].means(
            make_exhibit["no_shooting_testimony"]
        )

    def test_not_equal_different_speaker(self, make_exhibit):
        assert not (
            make_exhibit["no_shooting_different_witness_testimony"].means(
                make_exhibit["no_shooting_testimony"]
            )
        )

    def test_equal_complex_statement(self, make_exhibit):
        assert make_exhibit["relevant_murder_nested_swap_testimony"].means(
            make_exhibit["relevant_murder_testimony"]
        )

    def test_not_equal_complex_statement(self, make_exhibit):
        assert not (
            make_exhibit["relevant_murder_alice_craig_testimony"].means(
                make_exhibit["relevant_murder_testimony"]
            )
        )

    def test_not_equal_different_form(self, make_exhibit):
        assert not make_exhibit["shooting_affidavit"].means(
            make_exhibit["shooting_testimony"]
        )

    def test_explain_same_meaning(self, make_exhibit):
        explanation = make_exhibit["no_shooting_testimony"].explain_same_meaning(
            make_exhibit["no_shooting_entity_order_testimony"]
        )
        assert "<Alice> is like <Bob>, and <Bob> is like <Alice>" in str(explanation)


class TestExhibitsImplication:
    def test_implication(self, make_exhibit, caplog):
        caplog.set_level(logging.DEBUG)
        assert (
            make_exhibit["no_shooting_testimony"]
            > make_exhibit["no_shooting_witness_unknown_testimony"]
        )

    def test_no_implication_different_speaker(self, make_exhibit):
        assert (
            not make_exhibit["no_shooting_different_witness_testimony"]
            >= make_exhibit["no_shooting_testimony"]
        )

    def test_any_exhibit_implies_generic(self, make_exhibit):
        assert make_exhibit["reciprocal_testimony"] >= make_exhibit["generic_exhibit"]

    def test_exhibit_with_features_implies_featureless(self, make_exhibit):
        assert (
            make_exhibit["reciprocal_testimony"]
            >= make_exhibit["specific_but_featureless"]
        )

    def test_exhibit_no_implication_different_form(self, make_exhibit):
        assert (
            not make_exhibit["reciprocal_testimony"]
            >= make_exhibit["reciprocal_declaration"]
        )
        assert (
            not make_exhibit["reciprocal_testimony"]
            == make_exhibit["reciprocal_declaration"]
        )

    def test_implication_more_specific_testimony(self, make_exhibit):
        assert (
            make_exhibit["reciprocal_testimony_specific"]
            > make_exhibit["reciprocal_testimony"]
        )

    def test_implication_present_and_absent_testimony(self, make_exhibit):
        assert not (
            make_exhibit["reciprocal_testimony_specific_absent"]
            > make_exhibit["reciprocal_testimony"]
        )

    def test_absent_implies_more_specific_absent(self, make_exhibit):
        assert (
            make_exhibit["reciprocal_testimony_absent"]
            > make_exhibit["reciprocal_testimony_specific_absent"]
        )

    def test_absent_does_not_imply_less_specific_absent(self, make_exhibit):
        assert not (
            make_exhibit["reciprocal_testimony_specific_absent"]
            > make_exhibit["reciprocal_testimony_absent"]
        )

    def test_implication_fact_with_quantity(self, make_exhibit):
        specific_testimony = make_exhibit["large_weight_testimony"]
        vague_testimony = make_exhibit["small_weight_testimony"]
        assert specific_testimony >= vague_testimony

    def test_implication_fact_with_quantity(self, make_fact_about_exhibit):
        specific_testimony_reliable = make_fact_about_exhibit["f_reliable_large_weight"]
        vague_testimony_reliable = make_fact_about_exhibit["f_reliable_small_weight"]
        assert specific_testimony_reliable >= vague_testimony_reliable


class TestExhibitsContradiction:
    def test_conflicting_exhibits_not_contradictory(self, make_exhibit):
        assert not make_exhibit["shooting_testimony"].contradicts(
            make_exhibit["no_shooting_testimony"]
        )

    def test_absent_contradicts_same_present(self, make_exhibit):
        assert make_exhibit["no_shooting_witness_unknown_absent_testimony"].contradicts(
            make_exhibit["no_shooting_witness_unknown_testimony"]
        )

    def test_present_contradicts_same_absent(self, make_exhibit):
        assert make_exhibit["no_shooting_witness_unknown_absent_testimony"].contradicts(
            make_exhibit["no_shooting_witness_unknown_testimony"]
        )

    def test_more_specific_contradicts_absent(self, make_exhibit):
        assert make_exhibit["reciprocal_testimony_absent"].contradicts(
            make_exhibit["reciprocal_testimony_specific"]
        )
        assert make_exhibit["reciprocal_testimony_specific"].contradicts(
            make_exhibit["reciprocal_testimony_absent"]
        )

    def test_no_contradiction_with_factor_subclass(self, make_exhibit, watt_factor):
        assert not make_exhibit["shooting_testimony"].contradicts(watt_factor["f4"])
        assert not watt_factor["f4"].contradicts(make_exhibit["shooting_testimony"])



from copy import deepcopy
from datetime import date
import datetime

import pytest

from authorityspoke.decisions import Decision, Opinion


class TestDecision:
    def test_decision_string(self, make_decision):
        decision = make_decision["cardenas"]
        assert str(decision) == "People v. Cardenas, 31 Cal. 3d 897 (1982-07-08)"

    def test_decision_no_opinions(self):
        decision = Decision(date(2000, 2, 2))
        assert decision.majority is None
        assert not decision.implied_by(None)

    def test_posit_holdings(self, make_decision, make_holding):
        """
        "Positing" Holdings causes the Holdings to be assigned to the Decision's
        majority Opinion.
        """

        watt = make_decision["watt"]
        watt.posit([make_holding["h1"], make_holding["h2"]])
        assert watt.majority.holdings[-1] == make_holding["h2"]

    def test_need_opinion_to_posit_holding(self, make_holding):
        decision = Decision(date=(datetime.date(1900, 1, 1)))
        with pytest.raises(AttributeError):
            decision.posit(make_holding["h1"])


class TestImplication:
    def test_implication_of_decision_with_one_of_same_holdings(
        self, make_decision, make_decision_with_holding
    ):
        oracle = make_decision["oracle"]
        oracle_with_holdings = make_decision_with_holding["oracle"]
        oracle.posit(oracle_with_holdings.holdings[0])
        assert len(oracle.holdings) == 1
        assert len(oracle_with_holdings.holdings) > 10
        assert oracle_with_holdings >= oracle

    def test_decision_implied_by_holding(self, make_decision, make_holding):
        decision = make_decision["watt"]
        holding = make_holding["h1"]
        decision.posit(holding)
        assert decision.implied_by(holding)

    def test_decision_explain_implication(
        self, make_decision_with_holding, make_holding
    ):
        decision = make_decision_with_holding["watt"]
        holding = decision.holdings[0]
        explanation = decision.explain_implication(holding)
        assert explanation

    def test_decision_no_explanation_implication(self, make_decision_with_holding):
        watt = make_decision_with_holding["watt"]
        oracle = make_decision_with_holding["oracle"]
        explanation = watt.explain_implication(oracle)
        assert explanation is None

    def test_typeerror_to_compare_with_factor(
        self, make_decision_with_holding, make_factor
    ):
        watt = make_decision_with_holding["watt"]
        factor = make_factor["f_no_shooting"]
        with pytest.raises(TypeError):
            watt.explain_implication(factor)

    def test_decision_implies_decision_without_holdings(
        self, make_decision_with_holding, make_decision
    ):
        oracle = make_decision_with_holding["oracle"]
        blank = make_decision["lotus"]
        explanation = oracle.explain_implication(blank)
        assert not explanation.reasons

    def test_no_holdings_of_blank_decision(self):
        blank = Decision(date=(2000, 1, 2))
        assert len(blank.holdings) == 0

    # @pytest.mark.skip(reason="slow")
    def test_decision_implies_its_opinion(self, make_decision_with_holding):
        cardenas = make_decision_with_holding["cardenas"]
        assert cardenas >= cardenas.majority
        assert cardenas > cardenas.majority

    # @pytest.mark.skip(reason="slow")
    def test_opinion_implies_its_decision(self, make_decision_with_holding):
        cardenas = make_decision_with_holding["cardenas"]
        assert cardenas.implied_by(cardenas.majority)

    def test_decision_implies_holding_and_rule(self, make_decision_with_holding):
        """Adding a new input makes the new Holding less specific than the original one."""

        oracle = make_decision_with_holding["oracle"]
        holding = deepcopy(oracle.holdings[1])
        new_inputs = holding.inputs + [oracle.holdings[0].inputs[0]]
        holding.set_inputs(new_inputs)
        assert oracle.implies(holding)
        assert oracle.implies(holding.rule)

    def test_decision_does_not_imply_procedure(
        self, make_procedure, make_decision_with_holding
    ):
        oracle = make_decision_with_holding["oracle"]
        assert not oracle.implies(make_procedure["c1"])

    def test_opinion_implies_decision(self, make_holding, make_opinion_with_holding):
        watt = make_opinion_with_holding["watt_majority"]
        decision = Decision(
            date=date(2000, 1, 1), opinions=Opinion(position="majority")
        )
        decision.posit(watt.holdings[0])
        assert watt.implies(decision)

    def test_rule_implied_by_decision(self, make_decision_with_holding):
        oracle = make_decision_with_holding["oracle"]
        explanation = oracle.explain_implication(oracle.holdings[0].rule)
        assert len(explanation.reasons) == 3

    def test_decision_not_implied_by_rule(self, make_decision_with_holding):
        oracle = make_decision_with_holding["oracle"]
        explanation = oracle.explain_implied_by(oracle.holdings[0].rule)
        assert explanation is None


class TestContradiction:
    def test_oracle_contradicts_lotus(self, make_decision_with_holding):
        """
        This is the same example from the "Introduction" notebook.
        """
        oracle = make_decision_with_holding["oracle"]
        lotus = make_decision_with_holding["lotus"]
        assert oracle.contradicts(lotus)

    def test_no_contradiction_explanations(self, make_decision_with_holding):
        oracle = make_decision_with_holding["oracle"]
        feist = make_decision_with_holding["feist"]
        explanation = oracle.explain_contradiction(feist)
        assert explanation is None

    def test_cannot_check_contradiction_of_str(self, make_decision_with_holding):
        oracle = make_decision_with_holding["oracle"]
        feist = make_decision_with_holding["feist"]
        with pytest.raises(TypeError):
            oracle.explain_contradiction("feist")

    def test_no_contradiction_with_plurality(self, make_decision_with_holding):
        oracle = make_decision_with_holding["oracle"]
        other = Decision(
            date=datetime.date(2020, 1, 1), opinions=[Opinion(position="plurality")]
        )
        assert not oracle.contradicts(other)

    def test_no_contradiction_of_majority_without_holdings(
        self, make_decision_with_holding
    ):
        oracle = make_decision_with_holding["oracle"]
        other = Decision(
            date=datetime.date(2020, 1, 1), opinions=[Opinion(position="majority")]
        )
        assert not oracle.contradicts(other)

    def test_decision_contradicts_holding(self, make_decision_with_holding):
        oracle = make_decision_with_holding["oracle"]
        holding = oracle.holdings[0].negated()
        assert oracle.contradicts(holding)


import os
from dotenv import load_dotenv

import pytest

from legislice.download import Client
from authorityspoke import Entity, Fact, Predicate, Holding


load_dotenv()
LEGISLICE_API_TOKEN = os.getenv("LEGISLICE_API_TOKEN")

CLIENT = Client(api_token=LEGISLICE_API_TOKEN)


class TestAddHoldings:
    """
    Test creating Holdings based on United States v. Harmon.

    No. 2019-0395 (D.D.C. 2020) (December 24th, 2020)
    """

    @pytest.mark.vcr
    def test_create_and_add_holdings(self):
        offense_statute = CLIENT.read("/us/usc/t18/s1960/a")
        no_license = Fact(
            "$business was licensed as a money transmitting business",
            truth=False,
            terms=Entity("Helix"),
        )
        operated = Fact(
            "$person operated $business as a business",
            terms=[Entity("Harmon"), Entity("Helix")],
        )
        transmitting = Fact(
            "$business was a money transmitting business", terms=Entity("Helix")
        )
        offense = Fact(
            "$person committed the offense of conducting an unlicensed money transmitting business",
            terms=Entity("Harmon"),
        )
        offense_holding = Holding.from_factors(
            inputs=[operated, transmitting, no_license],
            outputs=offense,
            enactments=offense_statute,
            universal=True,
        )
        definition_statute = CLIENT.read("/us/usc/t18/s1960/b/2")
        bitcoin = Fact(
            "$business transferred bitcoin on behalf of the public",
            terms=Entity("Helix"),
        )
        bitcoin_holding = Holding.from_factors(
            inputs=bitcoin,
            outputs=transmitting,
            enactments=definition_statute,
            universal=True,
        )
        assert "was a money transmitting business" in str(offense_holding.inputs)
        # The combined effect of the two holdings above is that if
        # a person operated a business that transferred bitcoin on behalf of
        # the public without a "money transmitting business" license,
        # the person is guilty of the offense.
        result = bitcoin_holding + offense_holding
        assert "<Helix> was a money transmitting business" not in str(result.inputs)
        assert "<Helix> was a money transmitting business" in str(result.outputs)
        assert "transferred bitcoin on behalf of the public" in str(result.inputs)



"""Function for generating specification from Marshmallow schemas."""

from apispec import APISpec
from apispec_oneofschema import MarshmallowPlugin

from authorityspoke.io.schemas_json import HoldingSchema


def make_spec() -> APISpec:
    """Generate specification for data used to create AuthoritySpoke objects."""
    holding_spec = APISpec(
        title="AuthoritySpoke Holding API Schema",
        version="0.3.0",
        openapi_version="3.0.2",
        info=dict(description="An interface for annotating judicial holdings"),
        plugins=[MarshmallowPlugin()],
    )

    holding_spec.components.schema("Holding", schema=HoldingSchema)

    return holding_spec


# spec = make_spec()



import pytest

from marshmallow.exceptions import ValidationError

from nettlesome.terms import ContextRegister
from nettlesome.entities import Entity
from nettlesome.statements import Statement

from authorityspoke.io import readers


class TestMakeEntities:
    def test_make_entity_from_str_without_mentioned(self):
        """
        This fails because it needs to look up the string factor_records
        in a "mentioned" list, but no "mentioned" parameter is given.
        """
        with pytest.raises(ValidationError):
            print(readers.read_factor(record="Bradley"))

    def test_conversion_to_generic(self, make_entity):
        e = make_entity
        assert e["motel_specific"].generic is False
        new = e["motel_specific"].make_generic()
        assert new.generic is True

    def test_repr_equal_after_make_generic(self, make_entity):
        """
        see the docstring for :meth:`Factor._import_to_mapping`
        for an explanation of what led to __repr__s being
        compared for equality instead of the underlying objects.
        """
        e = make_entity
        motel = e["motel"]
        motel_b = motel.make_generic()
        assert repr(motel) == repr(motel_b)

    def test_new_context(self, make_entity):

        changes = ContextRegister.from_lists(
            [make_entity["motel"], make_entity["watt"]],
            [Entity("Death Star"), Entity("Darth Vader")],
        )
        motel = make_entity["motel"]
        assert motel.new_context(changes) == changes.get_factor(make_entity["motel"])


class TestSameMeaning:
    def test_specific_to_generic_different_object(self, make_entity):
        e = make_entity
        motel = e["motel_specific"]
        motel_b = motel.make_generic()
        assert motel is not motel_b
        assert not motel == motel_b

    def test_equality_generic_entities(self, make_entity):
        e = make_entity
        assert e["motel"].means(e["trees"])
        assert not e["motel"] == e["trees"]

    def test_entity_does_not_mean_statement(self):
        entity = Entity("Bob")
        statement = Statement("$person loves ice cream", terms=entity)
        assert not entity.means(statement)
        assert not statement.means(entity)


class TestImplication:
    def test_implication_of_generic_entity(self, make_entity):
        assert make_entity["motel_specific"] > make_entity["trees"]

    def test_generic_entity_does_not_imply_specific_and_different(self, make_entity):
        assert not make_entity["motel_specific"] < make_entity["trees"]

    def test_implication_same_except_generic(self, make_entity):
        assert make_entity["motel_specific"] > make_entity["motel"]

    def test_generic_entity_does_not_imply_specific_and_same(self, make_entity):
        assert not make_entity["motel_specific"] < make_entity["motel"]

    def test_same_entity_not_ge(self, make_entity):
        assert not make_entity["motel"] > make_entity["motel"]

    def test_implication_subclass(self, make_entity):
        assert make_entity["tree_search_specific"] >= make_entity["motel"]
        assert make_entity["tree_search"] > make_entity["motel"]

    def test_plural_true(self, make_opinion_with_holding):
        """
        holding_feist.json has an entity with the name "Rural's telephone listings"
        and "plural": true
        """
        feist = make_opinion_with_holding["feist_majority"]
        assert any(entity.plural is True for entity in feist.generic_terms())

    def test_implies_concrete_with_same_name(self):
        concrete = Entity("Bob", generic=False)
        other = Entity("Bob", generic=False)
        assert concrete.implies(other)
        assert concrete >= other
        assert not concrete > other

    def test_implication_concrete_with_different_name(self):
        concrete = Entity("Bob", generic=False)
        generic = Entity("Barb")
        assert concrete.implies(generic)
        assert concrete > generic
        assert concrete >= generic

    def test_entity_does_not_imply_statement(self):
        entity = Entity("Bob")
        statement = Statement("$person loves ice cream", terms=entity)
        assert not entity.implies(statement)
        assert not statement.implies(entity)
        assert not entity >= statement
        assert not statement >= entity
        assert not entity > statement
        assert not statement > entity


class TestContradiction:
    def test_error_contradiction_with_non_factor(self, make_entity, make_predicate):
        with pytest.raises(TypeError):
            assert make_entity["trees"].contradicts(make_predicate["p3"])

    def test_no_contradiction_of_other_factor(self, make_entity, watt_factor):
        assert not make_entity["trees"].contradicts(make_entity["watt"])
        assert not make_entity["trees"].contradicts(watt_factor["f1"])



"""Functions for dumping AuthoritySpoke objects to external formats."""

from typing import Any, Dict

from authorityspoke.io import schemas_json


def to_dict(item: Any) -> Dict:
    """Serialize AuthoritySpoke object to dict."""
    schema = schemas_json.get_schema_for_item(item.__class__.__name__)
    return schema.dump(obj=item)


def to_json(item: Any) -> str:
    """Serialize AuthoritySpoke object to JSON string."""
    schema = schemas_json.get_schema_for_item(item.__class__.__name__)
    return schema.dumps(obj=item)



r""":class:`.Factor`\s used to support :class:`.Fact` findings."""
from dataclasses import dataclass, field

from typing import ClassVar, Iterator, List, Optional, Tuple

from anchorpoint import TextQuoteSelector

from nettlesome.terms import ContextRegister
from nettlesome.entities import Entity
from nettlesome.formatting import indented

from nettlesome.factors import Factor
from authorityspoke.facts import Fact


@dataclass()
class Exhibit(Factor):
    """
    A source of information for use in litigation.

    :param form:
        a term describing the category of exhibit. For example: testimony,
        declaration, document, or recording.

    :param statement:
        a fact assertion made via the exhibit. For instance, if the exhibit
        is a document, this parameter could refer to a statement printed
        on the document.

    :param statement_attribution:
        the :class:`.Entity` that the exhibit imputes the statement to. For
        instance, for a signed declaration, this would refer to the person
        whose signature appears on the declaration, regardless of any
        authentication concerns. The statement_attribution parameter may
        appear without the statement parameter, especially if the content
        of the statement is irrelevant.

    :param name:
        a string identifier for the exhibit

    :param absent:
        if True, indicates that no exhibit meeting the description exists
        in the litigation. If the exhibit has merely been rejected as
        evidence, use the absent attribute on an :class:`Evidence` object
        instead.

    :param generic:
        if True, indicates that the specific attributes of the exhibit
        are irrelevant in the context of the :class:`.Holding` where
        the exhibit is being referenced.

    .. note
        The form parameter may be replaced by a limited
        ontology of terms when sufficient example data is available.
    """

    form: Optional[str] = None
    statement: Optional[Fact] = None
    statement_attribution: Optional[Entity] = None
    name: Optional[str] = None
    absent: bool = False
    generic: bool = False
    anchors: List[TextQuoteSelector] = field(default_factory=list)
    context_factor_names: ClassVar[Tuple[str, ...]] = (
        "statement",
        "statement_attribution",
    )

    def _means_if_concrete(
        self, other: Factor, context: ContextRegister
    ) -> Iterator[ContextRegister]:
        if isinstance(other, self.__class__) and self.form == other.form:
            yield from super()._means_if_concrete(other, context)

    def _implies_if_concrete(
        self, other: Factor, context: Optional[ContextRegister] = None
    ) -> Iterator[ContextRegister]:
        if isinstance(other, self.__class__) and (
            self.form == other.form or other.form is None
        ):
            yield from super()._implies_if_concrete(other, context)

    def __str__(self):
        """Represent object as string without line breaks."""
        string = f'{("attributed to " + self.statement_attribution.short_string) if self.statement_attribution else ""}'
        if self.statement:
            string += ", asserting " + self.statement.short_string + ","
        string = super().__str__().format(string)
        return string.replace("exhibit", self.form or "exhibit").strip()

    @property
    def wrapped_string(self):
        text = ""
        if self.form:
            text += f"in the FORM {self.form}"
        if self.statement:
            text += "\n" + indented("WITH THE ASSERTION:")
            factor_text = indented(self.statement.wrapped_string, tabs=2)
            text += f"\n{str(factor_text)},"
        if self.statement_attribution:
            text += "\n" + indented(
                f"ATTRIBUTED TO {self.statement_attribution.wrapped_string}"
            )

        return super().__str__().format(text)


@dataclass()
class Evidence(Factor):
    """
    An :class:`Exhibit` admitted by a court to aid a factual determination.

    :param exhibit:
        the thing that is being used to aid a factual determination

    :param to_effect:
        the :class:`.Fact` finding that would be supported by the evidence.
        If the Fact object includes a non-null standard_of_proof attribute, it
        indicates that that the evidence would support a factual finding by
        that standard of proof.

    :param name:
        a string identifier

    :param absent:
        if True, indicates that no evidence meeting the description has been
        admitted, regardless of whether a corresponding :class:`Exhibit` has
        been presented

    :param generic:
        if True, indicates that the specific attributes of the evidence
        are irrelevant in the context of the :class:`.Holding` where
        the evidence is being referenced.
    """

    exhibit: Optional[Exhibit] = None
    to_effect: Optional[Fact] = None
    name: Optional[str] = None
    absent: bool = False
    generic: bool = False
    anchors: List[TextQuoteSelector] = field(default_factory=list)
    context_factor_names: ClassVar[Tuple[str, ...]] = ("exhibit", "to_effect")

    def __str__(self):
        string = (
            f'{("of " + self.exhibit.short_string + " ") if self.exhibit else ""}'
            + f'{("which supports " + self.to_effect.short_string) if self.to_effect else ""}'
        )
        return super().__str__().format(string).strip().replace("Evidence", "evidence")

    @property
    def wrapped_string(self):
        text = ""
        if self.exhibit:
            text += "\n" + indented("OF:")
            factor_text = indented(self.exhibit.wrapped_string, tabs=2)
            text += f"\n{str(factor_text)}"
        if self.to_effect:
            text += "\n" + indented("INDICATING:")
            factor_text = indented(self.to_effect.wrapped_string, tabs=2)
            text += f"\n{str(factor_text)}"
        return super().__str__().format(text).strip()
